/*
 * data.c - Values, variables, and operators
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <inttypes.h>
#include <stdlib.h>
#include <assert.h>

#include "dwarfx.h"

#include "util.h"
#include "array.h"
#include "error.h"
#include "sbug.h"
#include "data.h"


const struct dwx_type *void_type;
const struct dwx_type *int8_type;
const struct dwx_type *int16_type;
const struct dwx_type *int32_type;
const struct dwx_type *int64_type;
const struct dwx_type *uint8_type;
const struct dwx_type *uint16_type;
const struct dwx_type *uint32_type;
const struct dwx_type *uint64_type;

/*
 * Memory objects that identify types which are private to the script
 * processor, and don't exist in this form in C.
 */

struct dwx_type type_type;
struct dwx_type string_type;
struct dwx_type lvalue_type;
struct dwx_type array_type;


int private_type(const struct dwx_type *type)
{
    return type == &type_type || type == &string_type ||
      type == &lvalue_type || type == &array_type ||
      type == &process_type || type == &breakpoint_type ||
      type == &object_type;
}


struct value *new_value(const struct dwx_type *type)
{
    struct value *value;

    value = alloc_t(struct value);
    value->type = type;
    memset(&value->u,0,sizeof(value->u));
    value->ref = 1;
    return value;
}


struct value *get(struct value *value)
{
    if (value) value->ref++;
    return value;
}


void put(struct value *value)
{
    if (!value) return;
    assert(value->ref);
    if (!--value->ref) {
	if (value->type == &lvalue_type) /* do nothing */;
	else if (value->type == &array_type) put_array(value);
	else if (value->type == &type_type) /* do nothing */;
	else if (sbug_put(value)) /* do nothing else */;
	else {
	    if (value->u.owner) put(value->u.owner);
	    else free((void *) value->value);
	}
	free(value);
    }
}


void share(struct value *value,struct value *owner)
{
    assert(!private_type(value->type));
    value->u.owner = get(owner->u.owner ? owner->u.owner : owner);
}


struct value *int_value(uint32_t num)
{
    struct value *value;

    value = new_value(uint32_type);
    value->value = alloc_t(uint32_t);
    *(uint32_t *) value->value = num;
    return value;
}


uint32_t value_int(struct value *value)
{
    uint32_t res;

    if (sbug_value_int(value,&res)) return res;
    if (value && value->type == &object_type)
	value = dereference_object(value,1);
    if (!value) {
	put(value);
	return 0;
    }
    if (private_type(value->type) ||
      (value->type->kind != dwx_uint && value->type->kind != dwx_sint &&
      value->type->kind != dwx_ptr))
        clerrorf("invalid type (expected integer)");
    if (value->type->kind == dwx_ptr) res = *(uint32_t *) value->value;
    else switch (value->type->size) {
	    case 1:
		res = *(uint8_t *) value->value;
		break;
	    case 2:
		res = *(uint16_t *) value->value;
		break;
	    case 4:
		res = *(uint32_t *) value->value;
		break;
	    case 8:
		/* @@@ not yet supported */
	    default:
		abort();
	}
    put(value);
    return res;
}


uint32_t int_eval(const struct op *op)
{
    struct value *value;

    value = evaluate_lvalue(op);
    if (value && value->type == &lvalue_type)
        value = dereference_lvalue(value);
    return value_int(value);
}


struct value *string_value(const char *str)
{
    struct value *value;

    value = new_value(&string_type);
    value->value = str;
    return value;
}


const char *__value_string(struct value *value)
{
    if (!value) return "";
    if (value->type != &string_type)
	clerrorf("invalid type (expected string)");
    return value->value;
}


const char *value_string(struct value *value)
{
    const char *res;

    assert(!value || value->ref > 1);
    res = __value_string(value);
    put(value);
    return res;
}


const char *string_eval(const struct op *op)
{
    return value_string(evaluate(op));
}


struct value *type_value(const struct dwx_type *type)
{
    struct value *value;

    value = new_value(&type_type);
    value->value = type;
    return value;
}


const struct dwx_type *value_type(struct value *value)
{
    const struct dwx_type *res;

    if (value->type != &type_type) clerrorf("invalid type (expected type)");
    res = value->value;
    put(value);
    return res;
}


const struct dwx_type *type_eval(const struct op *op)
{
    return value_type(evaluate(op));
}


struct value *array_value(void)
{
    struct value *value;

    value = new_value(&array_type);
    value->value = NULL;
    return value;
}


struct value *lvalue_value(struct value **value)
{
    struct value *lvalue;

    lvalue = new_value(&lvalue_type);
    lvalue->value = value;
    return lvalue;
}


struct value *evaluate_lvalue(const struct op *op)
{
    return op->fn(op);
}


struct value *dereference_lvalue(struct value *lvalue)
{
    struct value *res;

    if (lvalue->type != &lvalue_type) clerrorf("lvalue expected");
    res = get(*(struct value **) lvalue->value);
    put(lvalue);
    return res;
}


struct value *evaluate(const struct op *op)
{
    struct value *value;

    value = op->fn(op);
    if (!value) return value;
    if (value->type == &lvalue_type) value = dereference_lvalue(value);
    if (value && value->type == &object_type)
	return dereference_object(value,1);
    return value;
}


void init_types(void)
{
    int8_type = dwx_basic_type(dwx_sint,1);
    int16_type = dwx_basic_type(dwx_sint,2);
    int32_type = dwx_basic_type(dwx_sint,4);
    int64_type = dwx_basic_type(dwx_sint,8);
    uint8_type = dwx_basic_type(dwx_uint,1);
    uint16_type = dwx_basic_type(dwx_uint,2);
    uint32_type = dwx_basic_type(dwx_uint,4);
    uint64_type = dwx_basic_type(dwx_uint,8);
}
