/*
 * debug.c - Debugger interface
 *
 * Written 2003 by Werner Almesberger, Caltech Netlab FAST project
 */

#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>

#include "dwarfx.h"
#include "ptrace.h"

#include "util.h"
#include "error.h"
#include "data.h"
#include "var.h"
#include "op.h"
#include "stmt.h"
#include "sbug.h"


extern void open_elf(const char *name);



static struct process {
    struct pt_process *proc;
    struct breakpoint *breakpoints;
    struct breakpoint *current_breakpoint;
	/* current breakpoint; NULL if none */
    struct object *objects;
    struct dwx_obj *objs;
    struct value *value;	/* points back to the value */
    struct process *next;
} *processes = NULL;

struct breakpoint {
    struct process *process; /* does NOT count as a reference;
				NULL if process is dead */
    const struct dwx_obj *obj;
    unsigned long addr;
    struct value *value;	/* points back to the value */
    struct breakpoint *next; /* only valid if process != NULL */
};

/*
 * Note: we cannot use dwx_obj directly, because de-referencing pointers
 * and such would yield objects unknown to dwarfx.
 *
 * Also note the "constant" is different from a regular value, because it may
 * be a pointer, which points to an object in the debugged process. Therefore,
 * the process must be known.
 */

enum object_kind {
    ok_none,		/* catch 0 */
    ok_process,		/* only process */
    ok_unit,		/* compilation unit */
    ok_location,	/* location in process */
    ok_var,		/* variable in process */
    ok_const,		/* local constant */
};

struct object {
    enum object_kind kind;
    struct process *process;
    union {
	struct {
	    const struct dwx_obj *obj;
	} unit;
	struct {
	    const struct dwx_obj *obj;
	    unsigned long addr;
	} loc;
	struct {
	    const struct dwx_type *type;
	    const struct dwx_loc *loc;
	    unsigned long offset;
	} var;
	struct {
	    const struct dwx_type *type;
	    const void *pos;
	    int malloced; /* data is malloc'ed */
	} constant;
    } u;
    struct object *next;
};


/* ----- Helper functions -------------------------------------------------- */


static int process_has_terminated(const struct process *process)
{
    return process->proc->terminated;
}


static void check_process_termination(const struct process *process)
{
    if (!process) clerrorf("referenced process no longer exists");
    if (process_has_terminated(process)) clerrorf("process has terminated");
}


static void destroy_object(struct object *object)
{
    if (object->process) {
	struct object **walk;

	for (walk = &object->process->objects; *walk != object;
	  walk = &(*walk)->next);
	*walk = object->next;
    }
    if (object->kind == ok_const && object->u.constant.malloced)
	free((void *) object->u.constant.pos);
    free(object);
}


/* ----- Object composition and context ------------------------------------ */


static struct value **current_process_ptr = NULL;
  /* points into $$; NULL if we haven't touched any processes yet  */


static struct process *__current_process(void)
{
    if (!current_process_ptr || !*current_process_ptr ||
      (*current_process_ptr)->type != &process_type)
	return NULL;
    return __value_process(*current_process_ptr);
}


static struct process *current_process(void)
{
    struct process *process;

    process = __current_process();
    if (!process) clerrorf("no current process");
    return process;
}


static struct breakpoint *lookup_breakpoint(const struct process *process,
  struct pt_break *bp)
{
    struct breakpoint *b;

    if (!bp) return NULL;
    for (b = process->breakpoints; b; b = b->next)
	if (b->addr == bp->addr) break;
    return b;
}


static void set_current_process(struct process *process)
{
    struct var *var;

    var = var_lookup("$");
    put(var->value);
    var->value = process ? get(process->value) : NULL;
    current_process_ptr = &var->value;
    if (process)
	process->current_breakpoint =
	  lookup_breakpoint(process,process->proc->stopped);
}


static struct value *complete_object(struct object *object,
  const struct dwx_obj *obj)
{
    switch (obj->kind) {
	case dwx_obj_unit:
	    object->kind = ok_unit;
	    object->u.unit.obj = obj;
	    break;
	case dwx_obj_fn:
	case dwx_obj_label:
	    object->kind = ok_location;
	    object->u.loc.obj = obj;
	    object->u.loc.addr = obj->low_pc;
	    break;
	case dwx_obj_var:
	    if (obj->u.var.constant) {
		object->kind = ok_const;
		object->u.constant.type = obj->u.var.type;
		object->u.constant.pos = obj->u.var.constant;
		object->u.constant.malloced = 0;
	    }
	    else {
		object->kind = ok_var;
		object->u.var.type = obj->u.var.type;
		object->u.var.loc = obj->u.var.loc;
		object->u.var.offset = 0;
	    }
	    break;
	default:
	    abort();
    }
    object->next = object->process->objects;
    object->process->objects = object;
    return object_value(object);
}


static const struct dwx_obj *try_local(const struct process *process,
  const char *name)
{
    const struct dwx_obj *local;

    if (!process->current_breakpoint) return NULL;
    local = process->current_breakpoint->obj;
    if (local->kind == dwx_obj_label) local = local->parent;
    assert(local->kind == dwx_obj_fn);
    return dwx_find_item(local->children,name);
}


static struct value *compose_object(struct process *process,
  const struct object *base,const char *tag,int tag_is_unit)
{
    const struct dwx_obj *obj;
    struct object *object;

    object = alloc_t(struct object);
    if (base) *object = *base; /* don't clone - we'll link to process later */
    else {
	object->kind = ok_process;
	object->process = NULL;
    }
    if (!object->process)
	object->process = process ? process : current_process();
    else if (process) clerrorf("process is already selected");
    if (tag_is_unit) {
	if (object->kind != ok_process)
	   clerrorf("compilation unit is already selected");
	object->kind = ok_unit;
	if (tag) {
	    obj = dwx_find_unit(object->process->objs,tag);
	    if (!obj) clerrorf("no such compilation unit \"%s\"",tag);
	}
    }
    else {
	switch (object->kind) {
	    case ok_process:
		if (tag) {
		    obj = try_local(object->process,tag);
		    if (obj) {
			tag = NULL;
			break;
		    }
		}
		object->kind = ok_unit;
		object->u.unit.obj = object->process->objs;
		/* fall through */
	    case ok_unit:
		obj = object->u.unit.obj;
		break;
	    case ok_location:
		obj = object->u.loc.obj;
		break;
	    default:
		abort();
	}
	if (tag) {
	    obj = dwx_find_item(obj->children,tag);
	    if (!obj) clerrorf("no such object \"%s\"",tag);
	}
    }
    return complete_object(object,obj);
}


/* ----- Pointer and array conversions ------------------------------------- */


static struct object *clone_object(const struct object *object)
{
    struct object *new;

    new = alloc_t(struct object);
    *new = *object;
    if (new->kind == ok_const && new->u.constant.malloced) {
	int size;

	size = dwx_sizeof(object->u.constant.type);
	new->u.constant.pos = alloc(size);
	memcpy((void *) new->u.constant.pos,object->u.constant.pos,size);
    }
    new->next = object->process->objects;
    object->process->objects = new;
    return new;
}


static unsigned long var_data_location(const struct object *object)
{
    /* ia32-ism @@@ */
    return dwx_loc_here(object->u.var.loc,object->process->proc->regs.eip)+
      object->u.var.offset;
             
}


static struct value *fetch_pointer(struct value *value)
{
    struct object *object;

    object = (struct object *) __value_object(value);
    /*
     * @@@ Okay, this illustrates nicely that separating "objects" from
     * "values" is just plain dumb. Values should have an optional process
     * context (or, more general, some access methods), plus an indicator
     * of what to do with the data upon destruction.
     */
    switch (object->kind) {
	void *data;

	case ok_var:
	    object = clone_object(object);
	    object->kind = ok_const;
	    /*
	     * @@@ A bit of a type mixup: we need to allocate pointers as
	     * unsigned long (actually, should be some target_pointer_size),
	     * but all our numbers are uint32_t ...
	     */
	    data = alloc_t(unsigned long);
	    check_process_termination(object->process);
            pt_read(object->process->proc,var_data_location(object),data,
	      sizeof(unsigned long));
	    object->u.constant.pos = data;
	    object->u.constant.malloced = 1;
	    put(value);
	    return object_value(object);
	case ok_const:
	    return value;
	default:
	    abort();
    }
}


static struct value *to_pointer(struct value *value)
{
    struct object *object;
    const struct dwx_type *type;

    object = (struct object *) __value_object(value);
    if (object->kind != ok_var && object->kind != ok_const)
	clerrorf("wrong kind of object (data object expected)");
    type = dwx_strip_type(object->kind == ok_var ? object->u.var.type :
      object->u.constant.type);
    if (type->kind == dwx_ptr) return fetch_pointer(value);
    if (type->kind != dwx_array) clerrorf("pointer or array expected");
    /*
     * @@@ Okay, this illustrates nicely that separating "objects" from
     * "values" is just plain dumb. Values should have an optional process
     * context (or, more general, some access methods), plus an indicator
     * of what to do with the data upon destruction.
     */
    object = clone_object(object);
    type = dwx_add_ptr(type->u.array.type);
    /*
     * Like fetch_pointer, but avoid allocating temporary value
     */
    switch (object->kind) {
	unsigned long *data;

	case ok_var:
	    object->kind = ok_const;
	    data = alloc_t(unsigned long);
	    *data = var_data_location(object);
#if 0
	    check_process_termination(object->process);
	    
            pt_read(object->process->proc,var_data_location(object),data,
	      sizeof(unsigned long));
#endif
	    object->u.constant.pos = data;
	    object->u.constant.malloced = 1;
	    /* fall through */
	case ok_const:
	    object->u.constant.type = type;
	    break;
	default:
	    abort();
    }
    put(value);
    return object_value(object);
}


/* ----- Values ------------------------------------------------------------ */


struct dwx_type process_type;
struct dwx_type breakpoint_type;
struct dwx_type object_type;


struct value *process_value(struct process *process)
{
    struct value *value;

    value = new_value(&process_type);
    value->value = process;
    process->value = value;
    return value;
}


struct process *__value_process(struct value *value)
{
    if (!value || value->type != &process_type)
        clerrorf("invalid type (expected process)");
    return (struct process *) value->value;
}


struct value *breakpoint_value(struct breakpoint *breakpoint)
{
    struct value *value;

    value = new_value(&breakpoint_type);
    value->value = breakpoint;
    breakpoint->value = value;
    return value;
}


struct breakpoint *__value_breakpoint(struct value *value)
{
    if (!value || value->type != &breakpoint_type)
        clerrorf("invalid type (expected breakpoint)");
    return (struct breakpoint *) value->value;
}


struct value *object_value(struct object *obj)
{
    struct value *value;

    value = new_value(&object_type);
    value->value = obj;
    return value;
}


const struct object *__value_object(struct value *value)
{
    if (!value || value->type != &object_type)
        clerrorf("invalid type (expected object)");
    return value->value;
}


int sbug_put(struct value *value)
{
    if (value->type == &process_type) {
	struct process *process = (struct process *) value->value;
	struct process **walk;
	struct breakpoint *bp;
	struct object *object;

	pt_terminate(process->proc);
	for (walk = &processes; *walk != process; walk = &(*walk)->next);
	*walk = process->next;
	for (bp = process->breakpoints; bp; bp = bp->next) {
	    assert(bp->process == process);
	    bp->process = NULL;
	}
	for (object = process->objects; object; object = object->next) {
	    assert(object->process == process);
	    object->process = NULL;
	}
	free(process);
	return 1;
    }
    if (value->type == &breakpoint_type) {
	struct breakpoint *bp = (struct breakpoint *) value->value;

	if (bp->process) {
	    struct breakpoint **walk;

	    if (!process_has_terminated(bp->process))
		pt_unbreak(bp->process->proc,bp->addr);
	    for (walk = &bp->process->breakpoints; *walk != bp;
	      walk = &(*walk)->next);
	    *walk = bp->next;
	    if (bp->process->current_breakpoint == bp)
		bp->process->current_breakpoint = NULL;
	}
	free(bp);
	return 1;
    }
    if (value->type == &object_type) {
	struct object *object = (struct object *) value->value;

	destroy_object(object);
	return 1;
    }
    return 0;
}


int sbug_value_int(struct value *value,uint32_t *res)
{
    const struct object *object;
    const struct dwx_type *type;

    if (!value || value->type != &object_type) return 0;
    object = __value_object(value);
    switch (object->kind) {
	case ok_var:
	    type = object->u.var.type;
	    break;
	case ok_const:
	    type = object->u.constant.type;
	    break;
	default:
	    return 0;
    }
    type = dwx_strip_type(type);
    if (type->kind != dwx_ptr && type->kind != dwx_array) return 0;
    value = to_pointer(value);
    *res = *(unsigned long *) __value_object(value)->u.constant.pos;
    put(value);
    return 1;
}


/* ----- Operators: run ---------------------------------------------------- */


static int need_atexit = 1;


static void kill_processes(void)
{
    struct process *process;

    for (process = processes; process; process = process->next)
	pt_terminate(process->proc);
}


static char *lookup_program(char *name)
{
    const char *path,*pos,*colon;
    int len;

    if (strchr(name,'/')) return stralloc(name);
    path = getenv("PATH");
    if (!path) goto not_found;
    len = strlen(name);
    for (pos = path; *pos; pos = colon+1) {
	const char *end;
	char *tmp;

	colon = strchr(pos,':');
	if (colon) end = colon;
	else {
	    end = strchr(pos,0);
	    colon = end-1;
	}
	tmp = alloc(end-pos+len+2);
	    /* +2: / and \0 */
	memcpy(tmp,pos,end-pos);
	tmp[end-pos] = '/';
	strcpy(tmp+(end-pos)+1,name);
	if (!access(tmp,R_OK | X_OK)) return tmp;
	free(tmp);
    }
    /* fall through */

not_found:
    clerrorf("%s: command not found",name);
}


static struct dwx_obj *load_symbols(const char *path)
{
    struct dwx_obj *objs = NULL,**link = &objs;

    open_elf(path);
    while (1) {
	const void *p;

	p = dwarf_comp_unit();
	if (!p) break;
	*link = dwx_read_objects(p);
	link = &(*link)->next;
    }
    return objs;
}


static struct process *launch_process(const char *path,char **argv)
{
    struct process *process;

    process = alloc_t(struct process);
	/* allocate before fork, so we're sure it works */
    process->proc = pt_fork();
    if (process->proc) {
	process->current_breakpoint = NULL;
	process->breakpoints = NULL;
	process->objects = NULL;
	process->next = processes;
	processes = process;
	if (need_atexit) {
	    atexit(kill_processes);
	    need_atexit = 0;
	}
	return process;
    }
    else {
	if (execvp(path,argv) < 0) {
	    perror("execvp");
	    exit(1);
	}
	/* can't happen */
	exit(0);
    }
}


struct value *op_run(const struct op *self)
{
    struct arg_list *arg;
    int args;
    struct value *value0,*value;
    char **argv;
    char *path;
    struct dwx_obj *objs;
    struct process *process;

    args = 0;
    for (arg = self->u.run_args.args; arg; arg = arg->next)
	args++;
    argv = alloc((args+2)*sizeof(const char *));
	/* +2: argv[0], terminating NULL */
    value0 = evaluate(self->u.run_args.cmd);
    argv[0] = (char *) __value_string(value0);
    args = 1;
    for (arg = self->u.run_args.args; arg; arg = arg->next) {
	assert(!arg->value);
	arg->value = evaluate(arg->expr);
	argv[args] = (char *) __value_string(arg->value);
	args++;
    }
    argv[args] = NULL;
    path = lookup_program(*argv);
    objs = load_symbols(path);
    process = launch_process(path,argv);
    process->objs = objs;
    value = process_value(process);
    set_current_process(process);
    free(path);
    for (arg = self->u.run_args.args; arg; arg = arg->next) {
	put(arg->value);
	arg->value = NULL;
    }
    put(value0);
    free(argv);
    return value;
}


/* ----- Operators: break  ------------------------------------------------- */


struct value *op_breakpoint(const struct op *self)
{
    struct value *lvalue,*value;
    struct object *object;
    struct breakpoint *bp;

    lvalue = evaluate_lvalue(self->u.arg.a);
    if (lvalue && lvalue->type == &lvalue_type)
	lvalue = dereference_lvalue(lvalue);
    value = compose_object(NULL,__value_object(lvalue),NULL,0);
    object = (struct object *) __value_object(value);
    if (object->kind != ok_location)
	clerrorf("invalid object type (expected function or label)");
    check_process_termination(object->process);
    pt_break(object->process->proc,object->u.loc.addr);
    bp = alloc_t(struct breakpoint);
    bp->obj = object->u.loc.obj;
    bp->addr = object->u.loc.addr;
    bp->process = object->process;
    bp->next = object->process->breakpoints;
    object->process->breakpoints = bp;
    put(value);
    put(lvalue);
    return breakpoint_value(bp);
}


/* ----- Operators: object access ------------------------------------------ */


/*
 * Design issue: is this the right moment for binding an object to an address
 * space (process) ? Or should this be deferred to the time of evaluation ?
 * Normally, there isn't much of a difference, of course.
 */


struct value *op_object(const struct op *self)
{
    const char *name = string_eval(self->u.arg.a);

    if (__current_process()) {
	const struct dwx_obj *obj;

	obj = try_local(__current_process(),name);
	if (obj) {
	    struct object *object;

	    object = alloc_t(struct object);
	    object->process = __current_process();
	    return complete_object(object,obj);
	}
    }
    return compose_object(NULL,NULL,name,0);
}


/* ----- Operators: destructor --------------------------------------------- */


int destroy_sbug_op(const struct op *op)
{
    if (op->fn == op_run) {
	destroy_expr(op->u.run_args.cmd);
	destroy_arg_list(op->u.run_args.args);
	return 1;
    }
    return 0;
}


/* ----- Operators: op_index extension ------------------------------------- */


/*
 * Express foo[n] as *(foo+n)
 *
 * It is intentional that we ignore array bound here: it's fairly common to end
 * structures with an array of zero or one elements, and then allocating memory
 * for more. So the bounds would get in the way.
 */


int sbug_op_index(const struct op *self,struct value **res)
{
    struct op deref = {
	.fn = &op_dereference
    };
    struct op add = {
	.fn = &op_add
    };
    struct op get_value = {
	.fn = &op_const
    };
    struct value *value;

    value = evaluate_lvalue(self->u.arg.a);
    if (value->type != &object_type) goto bad;
    deref.u.arg.a = &add;
    get_value.u.value = value;
    add.u.arg.a = &get_value;
    add.u.arg.b = self->u.arg.b;
    *res = op_dereference(&deref);
    put(value);
    return 1;

bad:
    put(value); /* a bit pointless, but cleaner this way */
    return 0;
}


/* ----- Operators: op_member extension ------------------------------------ */


struct value *op_object_unit(const struct op *self)
{
    struct value *value,*res;

    value = evaluate(self->u.arg.a);
    if (value->type != &process_type)
	clerrorf("left-hand operand must be process");
    res = compose_object(__value_process(value),NULL,
      string_eval(self->u.arg.b),1);
    put(value);
    return res;
}


int sbug_op_member(const struct op *self,struct value **res)
{
    struct value *value;
    const struct object *object;
    struct object *new;
    const struct dwx_type *type;
    const char *tag;
    unsigned long offset = 0;

    value = evaluate_lvalue(self->u.arg.a);
    if (value->type == &lvalue_type) value = dereference_lvalue(value);
    tag = string_eval(self->u.arg.b);
    if (value->type == &process_type) {
	*res = compose_object(__value_process(value),NULL,tag,0);
	put(value);
	return 1;
    }
    if (value->type == &string_type) {
	struct value *tmp;

	tmp = compose_object(NULL,NULL,__value_string(value),1);
	*res = compose_object(NULL,__value_object(tmp),tag,0);
	put(value);
	put(tmp);
	return 1;
    }
    if (value->type != &object_type) goto bad;
    object = __value_object(value);
    switch (object->kind) {
	case ok_location:
	    if (object->u.loc.obj->kind == dwx_obj_label) return 0;
	    /* fall through */
	case ok_unit:
	    *res = compose_object(NULL,object,tag,0);
	    put(value);
	    return 1;
	case ok_var:
	    type = object->u.var.type;
	    break;
	case ok_const:
	    type = object->u.constant.type;
	    break;
	default:
	    goto bad;
    }
    if (type->kind != dwx_struct && type->kind != dwx_union) goto bad;
    new = clone_object(object);
    type = dwx_member(type,&offset,tag);
    switch (new->kind) {
	case ok_var:
	    new->u.var.type = type;
	    new->u.var.offset += offset;
	    break;
	case ok_const:
	    new->u.constant.type = type;
	    new->u.constant.pos += offset;
	    break;
	default:
	    abort();
    }
    put(value);
    *res = object_value(new);
    return 1;

bad:
    put(value); /* a bit pointless, but cleaner this way */
    return 0;
}


/* ----- Operators: current breakpoint ------------------------------------- */


struct value *op_current_bp(const struct op *self)
{
    struct process *curr;

    curr = __current_process();
    if (!curr || !curr->current_breakpoint) return NULL;
    return get(curr->current_breakpoint->value);
}


/* ----- Operators: op_eq and op_ne extension ------------------------------ */


static int sbug_relop(struct value *a,struct value *b,struct value **res,
  int not)
{
    int eq;

    if (a && b && a->type != b->type) return 0;
    if ((a && a->type == &process_type) ||
      (!a && b && b->type == &process_type))
	eq = (a ? __value_process(a) : NULL) ==
	  (b ? __value_process(b) : NULL);
    else if ((a && a->type == &breakpoint_type) ||
      (!a && b && b->type == &breakpoint_type))
	eq = (a ? __value_breakpoint(a) : NULL) ==
	  (b ? __value_breakpoint(b) : NULL);
    else return 0;
    *res = int_value(eq ^ not);
    put(a);
    put(b);
    return 1;
}


int sbug_op_eq(struct value *a,struct value *b,struct value **res)
{
    return sbug_relop(a,b,res,0);
}


int sbug_op_ne(struct value *a,struct value *b,struct value **res)
{
    return sbug_relop(a,b,res,1);
}


/* ----- Operators: & ------------------------------------------------------ */


struct value *op_address(const struct op *self)
{
    struct value *value;
    const struct object *object;
    const struct dwx_type *type,*stripped;
    struct object *new;

    value = evaluate_lvalue(self->u.arg.a);
    if (value && value->type == &lvalue_type)
	value = dereference_lvalue(value);
    if (!value || value->type != &object_type ||
      __value_object(value)->kind != dwx_obj_var)
	clerrorf("invalid operand (variable object expected)");
    object = __value_object(value);
    new = alloc_t(struct object);
    new->kind = ok_const;
    type = object->u.var.type;
    stripped = dwx_strip_type(type);
    if (stripped && stripped->kind == dwx_array) type = type->u.array.type;
    new->u.constant.type = dwx_add_ptr(type);
    new->u.constant.pos = alloc_t(unsigned long);
    *(unsigned long *) new->u.constant.pos = var_data_location(object);
    new->u.constant.malloced = 1;
    new->process = object->process;
    new->next = new->process->objects;
    new->process->objects = new;
    put(value);
    return object_value(new);
}


/* ----- Operators: * ------------------------------------------------------ */


static uint8_t zero_addr_expr[] = { DW_OP_lit0 };

static struct dwx_loc zero_addr = {
    .start = 0,
    .end = ~0U,
    .expr = zero_addr_expr,
    .length = sizeof(zero_addr_expr),
    .next = NULL,
};


struct value *op_dereference(const struct op *self)
{
    struct value *value;
    const struct object *object;
    const struct dwx_type *type;
    struct object *new;

    value = evaluate_lvalue(self->u.arg.a);
    if (value && value->type == &lvalue_type)
	value = dereference_lvalue(value);
    if (!value || value->type != &object_type)
	clerrorf("invalid operand type (data object expected)");
    /*
     * @@@ ... or defer process check until we really can't delay it any
     * further ? (i.e. only do it for ok_var)
     */
    value = to_pointer(value);
    object = __value_object(value);
    check_process_termination(object->process);
    switch (object->kind) {
	case ok_const:
	    type = object->u.constant.type;
	    break;
	case ok_var:
	    type = object->u.var.type;
	    break;
	default:
	    clerrorf("invalid operand (data object expected)");
    }
    type = dwx_strip_type(type);
    if (type->kind != dwx_ptr)
	clerrorf("invalid operand type (pointer expected)");
    new = alloc_t(struct object);
    new->kind = ok_var;
    new->u.var.type = type->u.modifier.type;
    new->u.var.loc = &zero_addr;
    if (object->kind == ok_const)
	new->u.var.offset = *(unsigned long *) object->u.constant.pos;
    else pt_read(object->process->proc,var_data_location(object),
	  &new->u.var.offset,sizeof(unsigned long));
    new->process = object->process;
    new->next = new->process->objects;
    new->process->objects = new;
    put(value);
    return object_value(new);
}


/* ----- Operators: -> ----------------------------------------------------- */


struct value *op_ptr(const struct op *self)
{
    /*
     * Not static, since we may recurse
     */
    struct op deref = {
	.fn = &op_dereference
    };
    struct op member = {
	.fn = &op_member,
    };

    deref.u.arg.a = self->u.arg.a;
    member.u.arg.a = &deref;
    member.u.arg.b = self->u.arg.b;
    return op_member(&member);
}


/* ----- Operators: + extension -------------------------------------------- */


static struct value *advance_pointer(struct value *ptr,int index)
{
    struct value *res;
    const struct object *object;

    object = __value_object(ptr);
    assert(object->kind == ok_const &&
      object->u.constant.type->kind == dwx_ptr);
    res = object_value(clone_object(object));
    object = __value_object(res);
    *(unsigned long *) object->u.constant.pos +=
      index*dwx_sizeof(object->u.constant.type->u.modifier.type);
    put(ptr);
    return res;
}


static int is_pointer(struct value *value)
{
    const struct object *object;
    const struct dwx_type *type;

    if (!value || value->type != &object_type) return 0;
    object = __value_object(value);
    switch (object->kind) {
	case ok_var:
	    type = object->u.var.type;
	    break;
	case ok_const:
	    type = object->u.constant.type;
	    break;
	default:
	    return 0;
    }
    type = dwx_strip_type(type);
    return type->kind == dwx_ptr || type->kind == dwx_array;
}


int sbug_op_add(struct value *a,struct value *b,struct value **res)
{
    struct value *ptr;
    int index;

    if (is_pointer(a)) {
	ptr = to_pointer(a);
	index = value_int(b);
    }
    else if (is_pointer(b)) {
	ptr = to_pointer(b);
	index = value_int(a);
    }
    else return 0;
    *res = advance_pointer(ptr,index);
    return 1;
}


/* ----- Operators: - extension -------------------------------------------- */


static struct value *ptr_diff(struct value *a,struct value *b)
{
    const struct object *oa,*ob;
    struct value *res;

    a = to_pointer(a);
    b = to_pointer(b);
    oa = __value_object(a);
    ob = __value_object(b);
    if (!dwx_type_equal(oa->u.constant.type,ob->u.constant.type))
	clerrorf("incompatible types");
    res = int_value((*(unsigned long *) oa->u.constant.pos-
      *(unsigned long *) ob->u.constant.pos)/dwx_sizeof(oa->u.constant.type));
    put(a);
    put(b);
    return res;
}


int sbug_op_sub(struct value *a,struct value *b,struct value **res)
{
    if (is_pointer(a) && is_pointer(b)) {
	*res = ptr_diff(a,b);
	return 1;
    }
    if (!is_pointer(a)) {
	if (is_pointer(b))
	    clerrorf("cannot subtract pointer from non-pointer");
	return 0;
    }
    *res = advance_pointer(to_pointer(a),-value_int(b));
    return 1;
}


/* ----- Operators: assignment extension (rvalue) -------------------------- */


struct value *sbug_assign_hook(struct value *value)
{
    if (!value || value->type != &object_type) return value;
    return dereference_object(value,0);
}


/* ----- Operators: sizeof extension --------------------------------------- */


int sbug_op_sizeof_expr(struct value *value,struct value **res)
{
    const struct object *object;
    const struct dwx_type *type;

    if (!value || value->type != &object_type) return 0;
    object = __value_object(value);
    switch (object->kind) {
	case ok_var:
	    type = object->u.var.type;
	    break;
	case ok_const:
	    type = object->u.constant.type;
	    break;
	default:
	    clerrorf("cannot apply sizeof to non-data objects");
    }
    *res = int_value(dwx_sizeof(type));
    put(value);
    return 1;
}


/* ----- Statements -------------------------------------------------------- */


void stmt_continue_process(const struct stmt *self)
{
    struct value *value;
    struct pt_process *proc;
    struct process *process;

    if (!self->u.process.process) process = current_process();
    else {
	value = evaluate(self->u.process.process);
	process = __value_process(value);
    }
    set_current_process(NULL);
    check_process_termination(process);
    pt_run(process->proc);
    if (self->u.process.process) put(value);
    /*
     * @@@ Not very flexible ...
     */
    proc = pt_wait(1);
    for (process = processes; process->proc != proc; process = process->next);
    set_current_process(process);
}


int destroy_sbug_stmt(const struct stmt *stmt)
{
    if (stmt->fn == stmt_continue_process) {
	destroy_expr(stmt->u.process.process);
	return 1;
    }
    return 0;
}


/* ----- Data access ------------------------------------------------------- */


struct value *dereference_object(struct value *value,int to_data)
{
    const struct object *object;
    const struct dwx_type *type;
    int size;
    struct value *new;
    void *data;

    if (is_pointer(value)) value = to_pointer(value);
    object = __value_object(value);
    switch (object->kind) {
	case ok_unit:
	case ok_location:
	    return value;
	case ok_var:
	    type = object->u.var.type;
	    size = dwx_sizeof(type);
	    data = alloc(size);
	    check_process_termination(object->process);
	    pt_read(object->process->proc,var_data_location(object),data,size);
	    break;
	case ok_const:
	    if (!to_data) return value;
	    /*
	     * @@@ Optimize: create global owner of constant data, then use
	     * this as new->owner
	     */
	    type = object->u.constant.type;
	    size = dwx_sizeof(type);
	    data = alloc(size);
	    memcpy(data,object->u.constant.pos,size);
	    break;
	default:
	    abort();
    }
    if (to_data) {
	new = new_value(type);
	new->value = data;
    }
    else {
	struct object *new_object;

	new_object = alloc_t(struct object);
	new_object->kind = ok_const;
	new_object->u.constant.type = type;
	new_object->u.constant.pos = data;
	new_object->u.constant.malloced = 1;
	new_object->process = object->process;
	new_object->next = object->process->objects;
	new_object->process->objects = new_object;
	new = object_value(new_object);
    }
    put(value);
    return new;
}
