/*
 * io.c - Input/Output
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <assert.h>

#include "util.h"
#include "error.h"
#include "io.h"


static int fd = -1;
static const void *pending = NULL;
static int pending_size;


void in_open(const char *name)
{
    if (name) {
	fd = open(name,O_RDONLY);
	if (fd < 0) {
	    perror(name);
	    exit(1);
	}
    }
    else {
	/* stdin is used for the config script, so move this elsewhere */
	fd = dup(0);
	if (fd < 0) {
	    perror("dup");
	    exit(1);
	}
	if (close(0) < 0) {
	    perror("close");
	    exit(1);
	}
    }
}


static void trim_pending(int len)
{
    pending_size -= len;
    if (!pending_size) {
	free((void *) pending);
	pending = NULL;
    }
    else {
	void *tmp;

	tmp = alloc(pending_size);
	memcpy(tmp,pending+len,pending_size);
	free((void *) pending);
	pending = tmp;
    }
}


const void *in_read_n(int size)
{
    void *buf,*pos;
    int len = 0,got;

    if (fd == -1) return NULL;
    buf = pos = alloc(size);
    if (pending) {
	len = size > pending_size ? pending_size : size;
	memcpy(buf,pending,len);
	trim_pending(len);
	pos += len;
	size -= len;
    }
    if (!size) return buf;
    got = read(fd,pos,size);
    if (got == size) return buf;
    if (got < 0) clperror("read");
    if (len+got) in_return(buf,len+got);
    else free(buf);
    return NULL;
}


const void *in_read(int size)
{
    const void *p;

    p = in_read_n(size);
    if (p) return p;
    clerrorf("short read (need %d bytes)",size);
}


void in_skip(int size)
{
    off_t off;

    if (pending) {
	int len;

	len = size > pending_size ? pending_size : size;
	trim_pending(len);
	size -= len;
    }
    if (!size) return;
    off = lseek(fd,0,SEEK_CUR);
    if (off == (off_t) -1) {
	in_free(in_read(size));
	return;
    }
    if (lseek(fd,size,SEEK_CUR) == (off_t) -1) clperror("lseek");
}


void in_return(const void *data,int size)
{
    assert(!pending);
    if (size) {
	pending = data;
	pending_size = size;
    }
}


int in_eof(void)
{
    const void *p;

    if (pending) return 0;
    p = in_read_n(1);
    if (!p) return 1;
    in_return(p,1);
    return 0;
}


void in_free(const void *data)
{
    free((void *) data);
}


void io_cleanup(void)
{
    if (pending) free((void *) pending);
}
