/*
 * op.c - Operator construction and evaluation
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "dwarfx.h"

#include "util.h"
#include "data.h"
#include "array.h"
#include "error.h"
#include "io.h"
#include "sprintf.h"
#include "sbug.h"
#include "op.h"


/* ----- Evaluation functions ---------------------------------------------- */


struct value *op_const(const struct op *self)
{
    return get(self->u.value);
}


static struct value *op_get_var(const struct op *self)
{
    struct value **value = (struct value **) &self->u.var->value;

    return lvalue_value(value);
}


struct value *op_logical_or(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) || int_eval(self->u.arg.b));
}


struct value *op_logical_and(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) && int_eval(self->u.arg.b));
}


struct value *op_bit_or(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) | int_eval(self->u.arg.b));
}


struct value *op_bit_xor(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) ^ int_eval(self->u.arg.b));
}


struct value *op_bit_and(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) & int_eval(self->u.arg.b));
}


struct value *op_eq(const struct op *self)
{
    struct value *a,*b,*res;

    a = evaluate(self->u.arg.a);
    b = evaluate(self->u.arg.b);
    if (sbug_op_eq(a,b,&res)) return res;
    return int_value(value_int(a) == value_int(b));
}


struct value *op_ne(const struct op *self)
{
    struct value *a,*b,*res;

    a = evaluate(self->u.arg.a);
    b = evaluate(self->u.arg.b);
    if (sbug_op_ne(a,b,&res)) return res;
    return int_value(value_int(a) != value_int(b));
}


struct value *op_lt(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) < int_eval(self->u.arg.b));
}


struct value *op_gt(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) > int_eval(self->u.arg.b));
}


struct value *op_le(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) <=int_eval(self->u.arg.b));
}


struct value *op_ge(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) >= int_eval(self->u.arg.b));
}


struct value *op_shl(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) << int_eval(self->u.arg.b));
}


struct value *op_shr(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) >> int_eval(self->u.arg.b));
}


struct value *op_add(const struct op *self)
{
    struct value *a,*b,*res;

    /*
     * Do incomplete evaluation first, and only evaluate real lvalues, so that
     * sbug's objects don't get de-referenced.
     */
    a = evaluate_lvalue(self->u.arg.a);
    if (a && a->type == &lvalue_type) a = dereference_lvalue(a);
    b = evaluate_lvalue(self->u.arg.b);
    if (b && b->type == &lvalue_type) b = dereference_lvalue(b);
    if (sbug_op_add(a,b,&res)) return res;
    if (a && a->type == &object_type) a = dereference_object(a,1);
    if (b && b->type == &object_type) b = dereference_object(b,1);
/*
 * @@@ just forcing everything to be a one-size-fits-all int is of course
 * nonsense. should do proper type propagation.
 */
    return int_value(value_int(a)+value_int(b));
}


struct value *op_sub(const struct op *self)
{
    struct value *a,*b,*res;
    int size;

    /*
     * Do incomplete evaluation first, and only evaluate real lvalues, so that
     * sbug's objects don't get de-referenced.
     */
    a = evaluate_lvalue(self->u.arg.a);
    if (a && a->type == &lvalue_type) a = dereference_lvalue(a);
    b = evaluate_lvalue(self->u.arg.b);
    if (b && b->type == &lvalue_type) b = dereference_lvalue(b);
    if (sbug_op_sub(a,b,&res)) return res;
    if (a && a->type == &object_type) a = dereference_object(a,1);
    if (b && b->type == &object_type) b = dereference_object(b,1);
    if (a && b && (private_type(a->type) || private_type(b->type)))
	clerrorf("invalid type for subtraction");
    if ((!a || a->type->kind != dwx_ptr) && b &&  b->type->kind == dwx_ptr)
	clerrorf("cannot subtract pointer from non-pointer");
    if (!a || !b || a->type->kind != dwx_ptr || b->type->kind != dwx_ptr)
	return int_value(value_int(a)-value_int(b));
    if (!dwx_type_equal(a->type,b->type))
	clerrorf("can only subtract identical types");
    if (!a->type->u.modifier.type) size = 1;
    else size = dwx_sizeof(a->type->u.modifier.type);
    res = int_value((*(unsigned long *) a->value-*(unsigned long *) b->value)/
      size);
    put(a);
    put(b);
    return res;
}


struct value *op_mult(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a)*int_eval(self->u.arg.b));
}


struct value *op_div(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a)/int_eval(self->u.arg.b));
}


struct value *op_mod(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) % int_eval(self->u.arg.b));
}


struct value *op_minus(const struct op *self)
{
    return int_value(-int_eval(self->u.arg.a));
}


struct value *op_logical_not(const struct op *self)
{
    return int_value(!int_eval(self->u.arg.a));
}


struct value *op_bit_not(const struct op *self)
{
    return int_value(~int_eval(self->u.arg.a));
}


/* ----- Constructors for constants ---------------------------------------- */


struct op *op_int(uint32_t num)
{
    struct op *op;

    op = alloc_t(struct op);
    op->fn = op_const;
    op->u.value = int_value(num);
    return op;
}


struct op *op_string(const char *str)
{
    struct op *op;

    op = alloc_t(struct op);
    op->fn = op_const;
    op->u.value = string_value(str);
    return op;
}


struct op *op_type(const struct dwx_type *type)
{
    struct op *op;

    op = alloc_t(struct op);
    op->fn = op_const;
    op->u.value = type_value(type);
    return op;
}


static struct op *op_value(struct value *value)
{
    struct op *op;

    op = alloc_t(struct op);
    op->fn = op_const;
    op->u.value = value;
    return op;
}


/* ----- Constructors for variables and types operators -------------------- */


struct op *op_var(const struct var *var)
{
    struct op *op;

    op = alloc_t(struct op);
    op->fn = op_get_var;
    op->u.var = var;
    return op;
}


static int size_of_value(struct value *value)
{
    int res;

    if (!value) clerrorf("undefined values have no size");
    if (value->type == &string_type) {
	res = strlen(value->value)+1;
	put(value);
	return res;
    }
    if (private_type(value->type)) clerrorf("invalid type");
    res = dwx_sizeof(value->type);
    put(value);
    return res;
}


struct value *op_cast(const struct op *self)
{
    const struct dwx_type *type;
    struct value *expr;

    type = type_eval(self->u.arg.a);
    expr = evaluate(self->u.arg.b);
    if (dwx_sizeof(type) > size_of_value(get(expr)))
	clerrorf("cannot cast to a larger type");
    expr->type = type;
    return expr;
}


struct value *op_sizeof_expr(const struct op *self)
{
    struct value *value,*res;

    value = evaluate_lvalue(self->u.arg.a);
    if (value && value->type == &lvalue_type)
	value = dereference_lvalue(value);
    if (sbug_op_sizeof_expr(value,&res)) return res;
    return int_value(size_of_value(value));
}


struct value *op_sizeof_type(const struct op *self)
{
    const struct dwx_type *type;

    type = type_eval(self->u.arg.a);
    if (private_type(type)) clerrorf("not a C type");
    return int_value(dwx_sizeof(type));
}


struct value *op_index(const struct op *self)
{
    struct value *value,*res;
    unsigned long offset;

    value = evaluate_lvalue(self->u.arg.a);
    if (value->type == &lvalue_type) {
        /* @@@ FIXME: $a[$a] should return "undefined value in index" */
	if (!*(struct value **) value->value)
	    *(struct value **) value->value = array_value();
	if ((*(struct value **) value->value)->type == &array_type) {
	    struct value **ptr;

	    ptr = (struct value **) value->value;
	    put(value);
	    return index_array(*ptr,evaluate(self->u.arg.b));
	}
	value = dereference_lvalue(value);
    }
    if (private_type(value->type) || value->type->kind != dwx_array) {
	/* a bit inefficient, but this helps keeping the interface clean */
	put(value);
	if (sbug_op_index(self,&res)) return res;
	clerrorf("trying to index a non-array");
    }
    offset = (unsigned long) value->value;
    res = new_value(dwx_index(value->type,&offset,int_eval(self->u.arg.b)));
    share(res,value);
    put(value);
    res->value = (const void *) offset;
    return res;
}


struct value *op_member(const struct op *self)
{
    struct value *value,*res;
    unsigned long offset;

    /*
     * Do incomplete evaluation first, and only evaluate real lvalues, so that
     * sbug's objects don't get de-referenced.
     */
    value = evaluate_lvalue(self->u.arg.a);
    if (value && value->type == &lvalue_type)
	value = dereference_lvalue(value);
    if (!value) clerrorf("trying to access member of undefined value");
    if (private_type(value->type) ||
      (value->type->kind != dwx_struct && value->type->kind != dwx_union)) {
	/* a bit inefficient, but this helps keeping the interface clean */
	put(value);
	if (sbug_op_member(self,&res)) return res;
	clerrorf("trying to access member a non-struct/union");
    }
    offset = (unsigned long) value->value;
    res = new_value(dwx_member(value->type,&offset,
      string_eval(self->u.arg.b)));
    share(res,value);
    put(value);
    res->value = (const void *) offset;
    return res;
}


struct value *op_add_array(const struct op *self)
{
    return type_value(
      dwx_add_array(type_eval(self->u.arg.a),int_eval(self->u.arg.b),-1));
}


struct value *op_add_fn(const struct op *self)
{
    return type_value(dwx_add_fn(type_eval(self->u.arg.a)));
}


struct value *op_add_ptr(const struct op *self)
{
    return type_value(dwx_add_ptr(type_eval(self->u.arg.a)));
}


/* ----- I/O operators ----------------------------------------------------- */


struct value *op_read(const struct op *self)
{
    struct value *value;
    const char *s,*p;
    int n;

    value = evaluate(self->u.arg.a);
    if (value && value->type == &type_type) {
	struct value *res;

	res = new_value(value_type(value));
	/* if using mmap, increment res->ref */
	res->value = in_read(dwx_sizeof(res->type));
	return res;
    }
    s = __value_string(value);
    n = strlen(s)+1;
    p = in_read_n(n);
    if (!p) {
	put(value);
	return int_value(0);
    }
    if (strncmp(p,s,n)) {
	put(value);
	in_return(p,n);
	return int_value(0);
    }
    put(value);
    in_free(p);
    return int_value(1);
}


struct value *op_more_data(const struct op *self)
{
    return int_value(!in_eof());
}


struct value *op_sprintf(const struct op *self)
{
    return string_value(common_sprintf(self->u.sprintf_args.fmt,
      self->u.sprintf_args.args));
}


/* ----- Conditional operator ---------------------------------------------- */


struct value *op_conditional(const struct op *self)
{
    return evaluate_lvalue(int_eval(self->u.arg.a) ? self->u.arg.b :
      self->u.arg.c);
}


/* ----- Assignment operator ----------------------------------------------- */


struct value *op_assignment(const struct op *self)
{
    struct value *lvalue,*rvalue;
    struct value **ptr;

    lvalue = evaluate_lvalue(self->u.arg.a);
    if (lvalue->type != &lvalue_type) clerrorf("lvalue expected");
    /*
     * Avoid unnecessarily converting sbug objects into the much simpler data
     * blocks of the basic language. That way, we can store pointers in
     * variables.
     */
    rvalue = evaluate_lvalue(self->u.arg.b);
    if (rvalue && rvalue->type == &lvalue_type)
	rvalue = dereference_lvalue(rvalue);
    rvalue = sbug_assign_hook(rvalue);
    rvalue = get(copy_if_array(rvalue));
    ptr = (struct value **) lvalue->value;
    put(lvalue);
    put(*ptr);
    return *ptr = rvalue;
}


static struct value *do_modify_assign(
  struct value *(*fn)(const struct op *self),
  struct value *lvalue,const struct op *rvalue)
{
    struct value **ptr;
    struct op op;

    ptr = (struct value **) lvalue->value;
    op.fn = fn;
    op.u.arg.a = op_value(dereference_lvalue(lvalue));
    op.u.arg.b = rvalue;
    put(*ptr);
    *ptr = evaluate_lvalue(&op);
    if (*ptr && (*ptr)->type == &lvalue_type) *ptr = dereference_lvalue(*ptr);
    destroy_expr(op.u.arg.a);
    return get(*ptr);
}


static struct value *modify_assign(const struct op *self,
  struct value *(*fn)(const struct op *self))
{
    struct value *lvalue;

    lvalue = evaluate_lvalue(self->u.arg.a);
    if (lvalue->type != &lvalue_type) clerrorf("lvalue expected");
    return do_modify_assign(fn,lvalue,self->u.arg.b);
}


struct value *op_add_assign(const struct op *self)
{
    return modify_assign(self,op_add);
}


struct value *op_sub_assign(const struct op *self)
{
    return modify_assign(self,op_sub);
}


struct value *op_mul_assign(const struct op *self)
{
    return modify_assign(self,op_mult);
}


struct value *op_div_assign(const struct op *self)
{
    return modify_assign(self,op_div);
}


struct value *op_mod_assign(const struct op *self)
{
    return modify_assign(self,op_mod);
}


struct value *op_shl_assign(const struct op *self)
{
    return modify_assign(self,op_shl);
}


struct value *op_shr_assign(const struct op *self)
{
    return modify_assign(self,op_shr);
}


struct value *op_and_assign(const struct op *self)
{
    return modify_assign(self,op_bit_and);
}


struct value *op_or_assign(const struct op *self)
{
    return modify_assign(self,op_bit_or);
}


struct value *op_xor_assign(const struct op *self)
{
    return modify_assign(self,op_bit_xor);
}


/* ----- Increment and decrement ------------------------------------------- */


static void xcrement(struct value *lvalue,int add)
{
    struct op op = {
	.fn = op_const
    };
    if (lvalue->type != &lvalue_type) clerrorf("lvalue expected");
    op.u.value = int_value(add);
    put(do_modify_assign(op_add,lvalue,&op));
    put(op.u.value);
}


struct value *op_pre_increment(const struct op *self)
{
    struct value *lvalue;

    lvalue = evaluate_lvalue(self->u.arg.a);
    xcrement(get(lvalue),1);
    return dereference_lvalue(lvalue);
}


struct value *op_pre_decrement(const struct op *self)
{
    struct value *lvalue;

    lvalue = evaluate_lvalue(self->u.arg.a);
    xcrement(get(lvalue),-1);
    return dereference_lvalue(lvalue);
}


struct value *op_post_increment(const struct op *self)
{
    struct value *lvalue,*value;

    lvalue = evaluate_lvalue(self->u.arg.a);
    value = dereference_lvalue(get(lvalue));
    xcrement(lvalue,1);
    return value;
}


struct value *op_post_decrement(const struct op *self)
{
    struct value *lvalue,*value;

    lvalue = evaluate_lvalue(self->u.arg.a);
    value = dereference_lvalue(get(lvalue));
    xcrement(lvalue,-1);
    return value;
}


/* ----- Constructors for regular operators -------------------------------- */


struct op *op_noop(struct value *(*fn)(const struct op *self))
{
    struct op *op;

    op = alloc_t(struct op);
    op->fn = fn;
    op->u.arg.a = op->u.arg.b = op->u.arg.c = NULL;
    return op;
}


struct op *op_unary(struct value *(*fn)(const struct op *self),
  const struct op *a)
{
    struct op *op;

    op = op_noop(fn);
    op->u.arg.a = a;
    return op;
}


struct op *op_binary(struct value *(*fn)(const struct op *self),
  const struct op *a,const struct op *b)
{
    struct op *op;

    op = op_unary(fn,a);
    op->u.arg.b = b;
    return op;
}


struct op *op_ternary(struct value *(*fn)(const struct op *self),
  const struct op *a,const struct op *b,const struct op *c)
{
    struct op *op;

    op = op_binary(fn,a,b);
    op->u.arg.c = c;
    return op;
}


/* ---- Destructors -------------------------------------------------------- */


void destroy_expr(const struct op *op)
{
    if (!op) return;   
    if (op->fn == op_const) put(op->u.value);
    else if (op->fn == op_get_var) /* do nothing */;
    else if (op->fn == op_sprintf) {
	destroy_expr(op->u.sprintf_args.fmt);
	destroy_arg_list(op->u.sprintf_args.args);
    }
    else if (destroy_sbug_op(op)) /* do nothing else */;
    else {
	destroy_expr(op->u.arg.a);
	destroy_expr(op->u.arg.b);
	destroy_expr(op->u.arg.c);
    }
    free((struct op *) op);
}


void destroy_arg_list(struct arg_list *list)
{
    while (list) {
	struct arg_list *next = list->next;

	destroy_expr(list->expr);
	assert(!list->value);
	free(list);
	list = next;
    }
}
