/*
 * op.h - Operator construction and evaluation
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */

#ifndef OP_H
#define OP_H

#include <inttypes.h>

#include "dwarfx.h"

struct op; /* darn recursive inclusion */

#include "data.h"
#include "var.h"


struct op {
    struct value *(*fn)(const struct op *self);
    union {
	struct {
	    const struct op *a,*b,*c;
	} arg;				/* all ops except the ones below */
	struct {
	    const struct op *fmt;
	    struct arg_list *args;
	} sprintf_args;			/* op_sprintf */
	struct {
	    const struct op *cmd;
	    struct arg_list *args;
	} run_args;			/* op_run */
	const struct var *var;		/* op_get_var */
	struct value *value;		/* op_const */
    } u;
};

struct arg_list {
    const struct op *expr;
    struct value *value; /* need to cache values */
    struct arg_list *next;
};


struct value *op_const(const struct op *self);

struct op *op_int(uint32_t num);
struct op *op_string(const char *str);

struct op *op_var(const struct var *var);
struct op *op_type(const struct dwx_type *type);

struct value *op_logical_or(const struct op *self);
struct value *op_logical_and(const struct op *self);
struct value *op_bit_or(const struct op *self);
struct value *op_bit_xor(const struct op *self);
struct value *op_bit_and(const struct op *self);
struct value *op_eq(const struct op *self);
struct value *op_ne(const struct op *self);
struct value *op_lt(const struct op *self);
struct value *op_gt(const struct op *self);
struct value *op_le(const struct op *self);
struct value *op_ge(const struct op *self);
struct value *op_shl(const struct op *self);
struct value *op_shr(const struct op *self);
struct value *op_add(const struct op *self);
struct value *op_sub(const struct op *self);
struct value *op_mult(const struct op *self);
struct value *op_div(const struct op *self);
struct value *op_mod(const struct op *self);
struct value *op_minus(const struct op *self);
struct value *op_logical_not(const struct op *self);
struct value *op_bit_not(const struct op *self);

struct value *op_pre_increment(const struct op *self);
struct value *op_pre_decrement(const struct op *self);
struct value *op_post_increment(const struct op *self);
struct value *op_post_decrement(const struct op *self);

struct value *op_read(const struct op *self);
struct value *op_more_data(const struct op *self);

struct value *op_sprintf(const struct op *self);

struct value *op_cast(const struct op *self);
struct value *op_sizeof_expr(const struct op *self);
struct value *op_sizeof_type(const struct op *self);
struct value *op_index(const struct op *self);
struct value *op_member(const struct op *self);

struct value *op_add_array(const struct op *self);
struct value *op_add_fn(const struct op *self);
struct value *op_add_ptr(const struct op *self);

struct value *op_assignment(const struct op *self);
struct value *op_add_assign(const struct op *self);
struct value *op_sub_assign(const struct op *self);
struct value *op_mul_assign(const struct op *self);
struct value *op_div_assign(const struct op *self);
struct value *op_mod_assign(const struct op *self);
struct value *op_shl_assign(const struct op *self);
struct value *op_shr_assign(const struct op *self);
struct value *op_and_assign(const struct op *self);
struct value *op_xor_assign(const struct op *self);
struct value *op_or_assign(const struct op *self);
struct value *op_conditional(const struct op *self);

struct op *op_noop(struct value *(*fn)(const struct op *self));
struct op *op_unary(struct value *(*fn)(const struct op *self),
  const struct op *a);
struct op *op_binary(struct value *(*fn)(const struct op *self),
  const struct op *a,const struct op *b);
struct op *op_ternary(struct value *(*fn)(const struct op *self),
  const struct op *a,const struct op *b,const struct op *c);

void destroy_expr(const struct op *op); /* the "const" is a lie */
void destroy_arg_list(struct arg_list *list);

#endif /* OP_H */
