%{
/*
 * script.y - Dump script language
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */

/*
 * This grammar is loosely based on the ANSI C grammar posted by Jeff Lee in
 * 1985. ftp://ftp.uu.net/usenet/net.sources/ansi.c.grammar.Z
 */

/*
 * The differences between this scripting language and C are described in
 * README.
 */


#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>

#include "dwarfx.h"

#include "util.h"
#include "data.h"
#include "var.h"
#include "op.h"
#include "stmt.h"
#include "sbug.h"


struct stmt *script;


/* ----- Type construction ------------------------------------------------- */


static struct op *curr_type; /* type under construction */


static struct posted_op {
    struct op *op;
    struct posted_op *next;
} *posted_ops = NULL;


static void apply_posted(const struct posted_op *end)
{
    while (posted_ops != end) {
	struct posted_op *next = posted_ops->next;

	posted_ops->op->u.arg.a = curr_type;
	curr_type = posted_ops->op;
	free(posted_ops);
	posted_ops = next;
    }
}


static void post_op(struct op *op)
{
    struct posted_op *new;

    new = alloc_t(struct posted_op);
    new->op = op;
    new->next = posted_ops;
    posted_ops = new;
}


/* ----- While loops ------------------------------------------------------- */


static struct while_loop {
    const struct stmt *loop;
    const char *label;
    struct while_loop *next;
} *while_stack = NULL;


static void push_while(const struct stmt *loop,const char *label)
{
    struct while_loop *entry;

    entry = alloc_t(struct while_loop);
    entry->loop = loop;
    entry->label = label;
    entry->next = while_stack;
    while_stack = entry;
}


static void pop_while(void)
{
    struct while_loop *next;

    next = while_stack->next;
    if (while_stack->label) free(while_stack->label);
    free(while_stack);
    while_stack = next;
}


static struct stmt *jump_statement(void (*fn)(struct stmt *self),char *label)
{
    struct stmt *stmt;

    stmt = new_stmt(fn);
    if (!label) {
	if (!while_stack) {
	    if (fn == stmt_break) yyerror("break without while");
	    else yyerror("continue without while");
	}
	stmt->u.jump_stmt.loop = while_stack->loop;
    }
    else {
	const struct while_loop *walk;

	for (walk = while_stack; walk; walk = walk->next)
	    if (walk->label && !strcmp(walk->label,label)) break;
	if (!walk) yyerror("no such label");
	    stmt->u.jump_stmt.loop = walk->loop;
	free(label);
    }
    return stmt;
}


/* ----- Helper functions -------------------------------------------------- */


static void sbug_only(void)
{
    if (!sbug) yyerror("syntax error");
}


%}


%union {
    uint32_t num;
    const char *str;
    struct var *var;
    const struct op *op;
    struct stmt *stmt;
    const struct without *without;
    const struct arg_list *args;
    const struct value *(*opfn)(const struct op *self);
    const struct posted_op *post;
    const struct dwx_obj *obj;
};


%token		TOK_UNSIGNED TOK_SIGNED
%token		TOK_VOID TOK_CHAR TOK_SHORT TOK_INT TOK_LONG
%token		TOK_INT8_T TOK_INT16_T TOK_INT32_T TOK_INT64_T
%token		TOK_UINT8_T TOK_UINT16_T TOK_UINT32_T TOK_UINT64_T
%token		TOK_STRUCT TOK_UNION TOK_ENUM

%token		TOK_PRINT TOK_PRINTF TOK_WITHOUT
%token		TOK_IF TOK_ELSE TOK_WHILE TOK_FOR TOK_BREAK TOK_CONTINUE
%token		TOK_EXIT TOK_DIE TOK_UNDEF

%token		TOK_INC TOK_DEC
%token		TOK_OR TOK_AND TOK_SHL TOK_SHR TOK_EQ TOK_NE TOK_LE TOK_GE
%token		TOK_SIZEOF TOK_READ TOK_SPRINTF
%token		ADD_ASSIGN SUB_ASSIGN MUL_ASSIGN DIV_ASSIGN MOD_ASSIGN
%token		AND_ASSIGN XOR_ASSIGN OR_ASSIGN
%token		SHR_ASSIGN SHL_ASSIGN

%token <num>	NUMBER ENUM
%token <str>	WORD STRING TYPEDEF_NAME
%token <var>	VARIABLE

%type <str>	opt_label_name opt_label

%type <op>	expression opt_expression
%type <op>	assignment_expression conditional_expression
%type <op>	logical_or_expression logical_and_expression
%type <op>	inclusive_or_expression exclusive_or_expression
%type <op>	and_expression
%type <op>	equality_expression relational_expression shift_expression
%type <op>	additive_expression multiplicative_expression
%type <op>	cast_expression unary_expression
%type <op>	postfix_expression primary_expression

%type <op>	type_name type_specifier type_name_or_expression

%type <stmt>	statements statement
%type <stmt>	block expression_statement undef print printf
%type <stmt>	if opt_else while for jump exit die

%type <without>	opt_without without_list without

%type <args>	arg_list arg opt_arg_list

%type <opfn>	assignment_operator

%token		TOK_RUN DOLLAR_BANG PTR_OP
%type <stmt>	sbug_statements
%type <op>	sbug_functions sbug_primaries

%nonassoc BENEATH_ELSE
%nonassoc TOK_ELSE

%%


/* ===== Statements and blocks ============================================= */


all:
    statements
	{
	    script = $1;
	    fix_statements(&script);
	}
    ;

statements:
	{
	    $$ = NULL;
	}
    | statement statements
	{
	    if (!$1) $$ = $2;
	    else {
		$$ = $1;
		$$->next = $2;
	    }
	}
    ;

statement:
    block
	{
	    $$ = $1;
	}
    | expression_statement
	{
	    $$ = $1;
	}
    | undef
	{
	    $$ = $1;
	}
    | print
	{
	    $$ = $1;
	}
    | printf
	{
	    $$ = $1;
	}
    | if
	{
	    $$ = $1;
	}
    | while
	{
	    $$ = $1;
	}
    | for
	{
	    $$ = $1;
	}
    | jump
	{
	    $$ = $1;
	}
    | exit
	{
	    $$ = $1;
	}
    | die
	{
	    $$ = $1;
	}
    | ';'
	{
	    $$ = NULL;
	}
    | sbug_statements
	{
	    $$ = $1;
	}
    ;

block:
    '{' statements '}'
	{
	    $$ = $2;
	}
    ;


/* ----- Assignment and undef ---------------------------------------------- */


expression_statement:
    expression ';'
	{
	    $$ = new_stmt(stmt_expression);
	    $$->u.expr.expr = $1;
	}
    ;


undef:
    TOK_UNDEF expression ';'
	{
	    $$ = new_stmt(stmt_undef);
	    $$->u.expr.expr = $2;
	}
    ;


/* ----- print ------------------------------------------------------------- */


print:
    TOK_PRINT arg_list opt_without ';'
	{
	    $$ = new_stmt(stmt_print);
	    $$->u.print.args = $2;
	    $$->u.print.without = $3;
	}
    ;

arg_list:
    arg
	{
	    $$ = $1;
	}
    | arg ',' arg_list
	{
	    $$ = $1;
	    $$->next = $3;
	}
    ;

arg:
    expression
	{
	    $$ = alloc_t(struct arg_list);
	    $$->expr = $1;
	    $$->value = NULL;
	    $$->next = NULL;
	}
    ;

opt_without:
	{
	    $$ = NULL;
	}
    | TOK_WITHOUT without_list
	{
	    $$ = $2;
	}
    ;

without_list:
    without
	{
	    $$ = $1;
	}
    | without ',' without_list
	{
	    $$ = $1;
	    $$->next = $3;
	}
    ;

without:
    expression
	{
	    $$ = alloc_t(struct without);
	    $$->expr = $1;
	    $$->next = NULL;
	}
    ;


/* ----- printf ------------------------------------------------------------ */


printf:
    TOK_PRINTF '(' expression opt_arg_list ')' ';'
	{
	    $$ = new_stmt(stmt_printf);
	    $$->u.printf.fmt = $3;
	    $$->u.printf.args = $4;
	}
    ;

opt_arg_list:
	{
	    $$ = NULL;
	}
    | ',' arg_list
	{
	    $$ = $2;
	}
    ;


/* ----- if ---------------------------------------------------------------- */


if:
    TOK_IF '(' expression ')' statement opt_else
	{
	    $$ = new_stmt(stmt_if);
	    $$->u.if_stmt.expr = $3;
	    $$->u.if_stmt.true = $5;
	    $$->u.if_stmt.false = $6;
	}
    ;


opt_else:
    %prec BENEATH_ELSE
	{
	    $$ = NULL;
	}
    | TOK_ELSE statement
	{
	    $$ = $2;
	}
    ;

/* ----- break and continue ------------------------------------------------ */


jump:
    TOK_BREAK opt_label_name ';'
	{
	    $$ = jump_statement(stmt_break,$2);
	}
    | TOK_CONTINUE opt_label_name ';'
	{
	    $$ = jump_statement(stmt_continue,$2);
	}
    ;

opt_label_name:
	{
	    $$ = NULL;
	}
    | WORD
	{
	    $$ = $1;
	}
    ;


/* ----- exit -------------------------------------------------------------- */


exit:
    TOK_EXIT opt_expression ';'
	{
	    $$ = new_stmt(stmt_exit);
	    $$->u.exit.expr = $2;
	}
    ;

opt_expression:
	{
	    $$ = NULL;
	}
    | expression
	{
	    $$ = $1;
	}
    ;


/* ----- die -------------------------------------------------------------- */


die:
    TOK_DIE opt_expression ';'
	{
	    $$ = new_stmt(stmt_die);
	    $$->u.die.expr = $2;
	}
    ;


/* ----- while ------------------------------------------------------------- */


while:
    opt_label TOK_WHILE '(' expression ')'
	{
	    $$ = new_stmt(stmt_while);
	    push_while($$,$1);
	}
      statement
	{
	    pop_while();
	    $$ = $<stmt>6;
	    $$->u.while_stmt.expr = $4;
	    $$->u.while_stmt.body = $7;
	}
    ;


opt_label:
	{
	    $$ = NULL;
	}
    | WORD ':'
	{
	    $$ = $1;
	}
    ;


/* ----- for --------------------------------------------------------------- */


for:
    opt_label TOK_FOR '(' opt_expression ';' opt_expression ';' opt_expression
      ')'
	{
	    $$ = new_stmt(stmt_for);
	    push_while($$,$1);
	}
      statement
	 {
	    pop_while();
	    $$ = $<stmt>10;
	    $$->u.for_stmt.init = $4;
	    $$->u.for_stmt.test = $6;
	    $$->u.for_stmt.next = $8;
	    $$->u.for_stmt.body = $11;
	}
    ;


/* ===== Types ============================================================= */


type_name:
    type_specifier
	{
	    $$ = $1;
	}
    | type_specifier
	{
	    curr_type = $1;
	}
       abstract_declarator
	{
	    $$ = curr_type;
	}
    ;


/* ----- Basic types ------------------------------------------------------- */


type_specifier:
    TOK_VOID
	{
	    $$ = op_type(NULL);
	}
    | TOK_CHAR
	{
	    $$ = op_type(int8_type);
	}
    | TOK_SIGNED TOK_CHAR
	{
	    $$ = op_type(int8_type);
	}
    | TOK_UNSIGNED TOK_CHAR
	{
	    $$ = op_type(uint8_type);
	}
    | TOK_SHORT
	{
	    $$ = op_type(int16_type);
	}
    | TOK_SIGNED TOK_SHORT
	{
	    $$ = op_type(int16_type);
	}
    | TOK_UNSIGNED TOK_SHORT
	{
	    $$ = op_type(uint16_type);
	}
    | TOK_INT
	{
	    $$ = op_type(int32_type);
	}
    | TOK_SIGNED TOK_INT
	{
	    $$ = op_type(int32_type);
	}
    | TOK_UNSIGNED TOK_INT
	{
	    $$ = op_type(uint32_type);
	}
    | TOK_LONG
	{
	    $$ = op_type(int32_type);
	}
    | TOK_SIGNED TOK_LONG
	{
	    $$ = op_type(int32_type);
	}
    | TOK_UNSIGNED TOK_LONG
	{
	    $$ = op_type(uint32_type);
	}
    | TOK_LONG TOK_LONG
	{
	    $$ = op_type(int64_type);
	}
    | TOK_SIGNED TOK_LONG TOK_LONG
	{
	    $$ = op_type(int64_type);
	}
    | TOK_UNSIGNED TOK_LONG TOK_LONG
	{
	    $$ = op_type(uint64_type);
	}
    | TOK_INT8_T
	{
	    $$ = op_type(int8_type);
	}
    | TOK_INT16_T
	{
	    $$ = op_type(int16_type);
	}
    | TOK_INT32_T
	{
	    $$ = op_type(int32_type);
	}
    | TOK_INT64_T
	{
	    $$ = op_type(int64_type);
	}
    | TOK_UINT8_T
	{
	    $$ = op_type(uint8_type);
	}
    | TOK_UINT16_T
	{
	    $$ = op_type(uint16_type);
	}
    | TOK_UINT32_T
	{
	    $$ = op_type(uint32_type);
	}
    | TOK_UINT64_T
	{
	    $$ = op_type(uint64_type);
	}
    | TOK_STRUCT WORD
	{
	    const struct dwx_type *type;

	    type = dwx_lookup_type(dwx_struct,$2,-1,0);
	    if (!type) yyerror("unknown struct");
	    free($2);
	    $$ = op_type(type);
	}
    | TOK_UNION WORD
	{
	    const struct dwx_type *type;

	    type = dwx_lookup_type(dwx_union,$2,-1,0);
	    if (!type) yyerror("unknown union");
	    free($2);
	    $$ = op_type(type);
	}
    | TOK_ENUM WORD
	{
	    const struct dwx_type *type;

	    type = dwx_lookup_type(dwx_enum,$2,-1,0);
	    if (!type) yyerror("unknown enum");
	    free($2);
	    $$ = op_type(type);
	}
    | TYPEDEF_NAME
	{
	    const struct dwx_type *type;

	    type = dwx_lookup_type(dwx_typedef,$1,-1,0);
	    if (!type) yyerror("unknown typedef");
	    free($1);
	    $$ = op_type(type);
	}
    ;


/* ----- Abstract declarators ---------------------------------------------- */


abstract_declarator:
    pointer
    |	{
	    $<post>$ = posted_ops;
	}
      abstract_declarator2
	{
	    apply_posted($<post>1);
	}
    | pointer
	{
	    $<post>$ = posted_ops;
	}
      abstract_declarator2
	{
	    apply_posted($<post>2);
	}
    ;

abstract_declarator2:
    '(' abstract_declarator ')'
    | '[' expression ']'
	{
	    post_op(op_binary(op_add_array,NULL,$2));
	}
    | abstract_declarator2 '[' expression ']'
	{
	    post_op(op_binary(op_add_array,NULL,$3));
	}
    | '(' ')'
	{
	    post_op(op_unary(op_add_fn,NULL));
	}
    | abstract_declarator2 '(' ')'
	{
	    post_op(op_unary(op_add_fn,NULL));
	}
    ;

pointer:
    '*'
	{
	    curr_type = op_unary(op_add_ptr,curr_type);
	}
    | '*' pointer
	{
	    curr_type = op_unary(op_add_ptr,curr_type);
	}
    ;


/* ===== Expressions ======================================================= */


expression:
    assignment_expression
	{
	    $$ = $1;
	}
    ;

assignment_expression:
    conditional_expression
	{
	    $$ = $1;
	}
    | unary_expression assignment_operator assignment_expression
	{
	    $$ = op_binary($2,$1,$3);
	}
    ;

assignment_operator:
    '='
	{
	    $$ = op_assignment;
	}
    | MUL_ASSIGN
	{
	    $$ = op_mul_assign;
	}
    | DIV_ASSIGN
	{
	    $$ = op_div_assign;
	}
    | MOD_ASSIGN
	{
	    $$ = op_mod_assign;
	}
    | ADD_ASSIGN
	{
	    $$ = op_add_assign;
	}
    | SUB_ASSIGN
	{
	    $$ = op_sub_assign;
	}
    | SHL_ASSIGN
	{
	    $$ = op_shl_assign;
	}
    | SHR_ASSIGN
	{
	    $$ = op_shr_assign;
	}
    | AND_ASSIGN
	{
	    $$ = op_and_assign;
	}
    | XOR_ASSIGN
	{
	    $$ = op_xor_assign;
	}
    | OR_ASSIGN
	{
	    $$ = op_or_assign;
	}
    ;

conditional_expression:
    logical_or_expression
	{
	    $$ = $1;
	}
    | logical_or_expression '?' logical_or_expression ':'
      conditional_expression
	{
	    $$ = op_ternary(op_conditional,$1,$3,$5);
	}
    ;

logical_or_expression:
    logical_and_expression
	{
	    $$ = $1;
	}
    | logical_or_expression TOK_OR logical_and_expression
	{
	    $$ = op_binary(op_logical_or,$1,$3);
	}
    ;

logical_and_expression:
    inclusive_or_expression
	{
	    $$ = $1;
	}
    | logical_and_expression TOK_AND inclusive_or_expression
	{
	    $$ = op_binary(op_logical_and,$1,$3);
	}
    ;

inclusive_or_expression:
    exclusive_or_expression
	{
	    $$ = $1;
	}
    | inclusive_or_expression '|' exclusive_or_expression
	{
	    $$ = op_binary(op_bit_or,$1,$3);
	}
    ;

exclusive_or_expression:
    and_expression
	{
	    $$ = $1;
	}
    | exclusive_or_expression '^' and_expression
	{
	    $$ = op_binary(op_bit_xor,$1,$3);
	}
    ;

and_expression:
    equality_expression
	{
	    $$ = $1;
	}
    | and_expression '&' equality_expression
	{
	    $$ = op_binary(op_bit_and,$1,$3);
	}
    ;

equality_expression:
    relational_expression
	{
	    $$ = $1;
	}
    | equality_expression TOK_EQ relational_expression
	{
	    $$ = op_binary(op_eq,$1,$3);
	}
    | equality_expression TOK_NE relational_expression
	{
	    $$ = op_binary(op_ne,$1,$3);
	}
    ;

relational_expression:
    shift_expression
	{
	    $$ = $1;
	}
    | relational_expression '<' shift_expression
	{
	    $$ = op_binary(op_lt,$1,$3);
	}
    | relational_expression '>' shift_expression
	{
	    $$ = op_binary(op_gt,$1,$3);
	}
    | relational_expression TOK_LE shift_expression
	{
	    $$ = op_binary(op_le,$1,$3);
	}
    | relational_expression TOK_GE shift_expression
	{
	    $$ = op_binary(op_ge,$1,$3);
	}
    ;

shift_expression:
    additive_expression
	{
	    $$ = $1;
	}
    | shift_expression TOK_SHL additive_expression
	{
	    $$ = op_binary(op_shl,$1,$3);
	}
    | shift_expression TOK_SHR additive_expression
	{
	    $$ = op_binary(op_shr,$1,$3);
	}
    ;

additive_expression:
    multiplicative_expression
	{
	    $$ = $1;
	}
    | additive_expression '+' multiplicative_expression
	{
	    $$ = op_binary(op_add,$1,$3);
	}
    | additive_expression '-' multiplicative_expression
	{
	    $$ = op_binary(op_sub,$1,$3);
	}
    ;

multiplicative_expression:
    cast_expression
	{
	    $$ = $1;
	}
    | multiplicative_expression '*' cast_expression
	{
	    $$ = op_binary(op_mult,$1,$3);
	}
    | multiplicative_expression '/' cast_expression
	{
	    $$ = op_binary(op_div,$1,$3);
	}
    | multiplicative_expression '%' cast_expression
	{
	    $$ = op_binary(op_mod,$1,$3);
	}
    ;

cast_expression:
    unary_expression
	{
	    $$ = $1;
	}
    | '(' type_name ')' cast_expression
	{
	    $$ = op_binary(op_cast,$2,$4);
	}
    ;

unary_expression:
    postfix_expression
	{
	    $$ = $1;
	}
    | TOK_INC unary_expression
	{
	    $$ = op_unary(op_pre_increment,$2);
	}
    | TOK_DEC unary_expression
	{
	    $$ = op_unary(op_pre_decrement,$2);
	}
    | '&' cast_expression
	{
	    sbug_only();
	    $$ = op_unary(op_address,$2);
	}
    | '*' cast_expression
	{
	    sbug_only();
	    $$ = op_unary(op_dereference,$2);
	}
    | '-' cast_expression
	{
	    $$ = op_unary(op_minus,$2);
	}
    | '+' cast_expression
	{
	    $$ = $2;
	}
    | '!' cast_expression
	{
	    $$ = op_unary(op_logical_not,$2);
	}
    | '~' cast_expression
	{
	    $$ = op_unary(op_bit_not,$2);
	}
    | TOK_SIZEOF unary_expression
	{
	    $$ = op_unary(op_sizeof_expr,$2);
	}
    | TOK_SIZEOF '(' type_name ')'
	{
	    $$ = op_unary(op_sizeof_type,$3);
	}
    ;

postfix_expression:
    primary_expression
	{
	    $$ = $1;
	}
    | postfix_expression '[' expression ']'
	{
	    $$ = op_binary(op_index,$1,$3);
	}
    | postfix_expression '.' WORD
	{
	    $$ = op_binary(op_member,$1,op_string($3));
	}
    | postfix_expression '.' STRING
	{
	    sbug_only();
	    $$ = op_binary(op_object_unit,$1,op_string($3));
	}
    | postfix_expression PTR_OP WORD
	{
	    sbug_only();
	    $$ = op_binary(op_ptr,$1,op_string($3));
	}
    | postfix_expression TOK_INC
	{
	    $$ = op_unary(op_post_increment,$1);
	}
    | postfix_expression TOK_DEC
	{
	    $$ = op_unary(op_post_decrement,$1);
	}
    | TOK_READ '(' type_name_or_expression ')'
	{
	    if (sbug) yyerror("syntax error");
	    $$ = op_unary(op_read,$3);
	}
    | TOK_READ
	{
	    if (sbug) yyerror("syntax error");
	    $$ = op_noop(op_more_data);
	}
    | TOK_SPRINTF '(' expression opt_arg_list ')'
	{
	    $$ = op_noop(op_sprintf);
	    $$->u.sprintf_args.fmt = $3;
	    $$->u.sprintf_args.args = $4;
	}
    | sbug_functions
	{
	     $$ = $1;
	}
    ;

type_name_or_expression:
    type_name
	{
	    $$ = $1;
	}
    | expression
	{
	    $$ = $1;
	}
    ;

primary_expression:
    VARIABLE
	{
	    $$ = op_var($1);
	}
    | NUMBER
	{
	    $$ = op_int($1);
	}
    | ENUM
	{
	    $$ = op_int($1);
	}
    | STRING
	{
	    $$ = op_string($1);
	}
    | '(' expression ')'
	{
	    $$ = $2;
	}
    | sbug_primaries
	{
	    $$ = $1;
	}
    ;


/* ===== Sbug-specific material ============================================ */


sbug_statements:
    TOK_CONTINUE '(' opt_expression ')' ';'
	{
	    sbug_only();
	    $$ = new_stmt(stmt_continue_process);
	    $$->u.process.process = $3;
	}
    ;

sbug_functions:
    TOK_BREAK '(' expression ')'
	{
	    sbug_only();
	    $$ = op_unary(op_breakpoint,$3);
	}
    | TOK_RUN '(' expression opt_arg_list ')'
	{
	    sbug_only();
	    $$ = op_noop(op_run);
	    $$->u.run_args.cmd = $3;
	    $$->u.run_args.args = $4;
	}
    ;

sbug_primaries:
    WORD
	{
	    sbug_only();
	    $$ = op_unary(op_object,op_string($1));
	}
    | DOLLAR_BANG
	{
	    $$ = op_noop(op_current_bp);
	}
    ;
