/*
 * dwarf.c - DWARF2 decoder
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */

#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "dwarfhelp.h"
#include "dwarf.h"
#include "util.h"

#ifndef LITTLE_ENDIAN
#define LITTLE_ENDIAN
#endif


int dwarf_children = 0;

extern int debug;

static const void *curr_debug_abbrev,*curr_debug_info,*curr_debug_info_end;
static const void *curr_debug_str;
static const void *curr_comp_unit,*curr_comp_unit_end;


/* ----- LEB128 encode/decode functions ------------------------------------ */

/*
 * Adapted from Appendix 4 of DWARF specification
 */


#ifdef UNUSED

void uLEB128_encode(void **buf,unsigned value)
{
    unsigned char *p = *buf;

    while (1) {
	*p = value & 0x7f;
	value >>= 7;
	if (!value) break;
	*p++ |= 0x80;
    }
    *buf = p+1;
}


void sLEB128_encode(void **buf,int value)
{
    unsigned char *p = *buf;
    int negative = value < 0;

    while (1) {
	*p = value & 0x7f;
	value >>= 7;
	if (negative) value |= 1 << (sizeof(int)*8-7);
	if ((!value && !(*p & 0x40)) || (value == -1 && (*p & 0x40))) break;
	*p++ |= 0x80;
    }
    *buf = p+1;
}

#endif


unsigned uLEB128_decode(const void **buf)
{
    const unsigned char *p = *buf;
    unsigned result = 0;
    int shift;

    shift = 0;
    while (1) {
	result |= (*p & 0x7f) << shift;
	if (!(*p++ & 0x80)) break;
	shift += 7;
    }
    *buf = p;
    return result;
}


int sLEB128_decode(const void **buf)
{
    const unsigned char *p = *buf;
    int result = 0;
    int shift;

    while (1) {
	result |= (*p & 0x7f) << shift;
	shift += 7;
	if (!(*p & 0x80)) break;
	p++;
    }
    if (shift < 8*sizeof(int) && (*p & 0x40))
	result |= -(1<< shift);
    *buf = p+1;
    return result;
}


/* ----- Value decoder ----------------------------------------------------- */


uint16_t read_16(const unsigned char *p)
{
#ifdef LITTLE_ENDIAN
    return p[0] | (p[1] << 8);
#else
    return (p[0] << 8) | p[1];
#endif
}


uint32_t read_32(const unsigned char *p)
{
#ifdef LITTLE_ENDIAN
    return p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24);
#else
    return (p[0] << 24) | (p[1] << 16) | (p[2] << 8) | p[3];
#endif
}


void *dwarf_decode_addr(enum DW_FORM form,const void *p)
{
    if (form != DW_FORM_addr) {
	fprintf(stderr,"unrecognized address form %u\n",form);
	exit(1);
    }
    return (void *) read_32(p);
}


const void *dwarf_decode_block(enum DW_FORM form,const void *p,int *length)
{
    int len;

    switch (form) {
	case DW_FORM_block1:
	    len = *(const uint8_t *) p;
	    p++;
	    break;
	case DW_FORM_block2:
	    len = read_16(p);
	    p += 2;
	    break;
	case DW_FORM_block4:
	    len = read_32(p);
	    p += 4;
	    break;
	case DW_FORM_block:
	    len = uLEB128_decode(&p);
	    break;
	default:
	    fprintf(stderr,"unrecognized block form %u\n",form);
	    exit(1);
    }
    if (debug) printf("block @%p+%d\n",p,len);
    if (length) *length = len;
    return p;
}


int dwarf_decode_sconst(enum DW_FORM form,const void *p)
{
    unsigned tmp;

    switch (form) {
	case DW_FORM_data1:
	    return *(const int8_t *) p;
	case DW_FORM_data2:
	    tmp = read_16(p);
	    if (tmp & 0x8000) tmp |= ~0 & ~0xffff;
	    return tmp;
	case DW_FORM_data4:
	    tmp = read_32(p);
	    if (tmp & 0x80000000) tmp |= ~0 & ~0xffffffff;
	    return tmp;
	/* @@@ we don't support data8 yet */
	case DW_FORM_sdata:
	    return sLEB128_decode(&p);
	default:
	    fprintf(stderr,"unrecognized signed constant form %u\n",form);
	    exit(1);
    }
}


unsigned dwarf_decode_uconst(enum DW_FORM form,const void *p)
{
    switch (form) {
	case DW_FORM_data1:
	    return *(const uint8_t *) p;
	case DW_FORM_data2:
	    return read_16(p);
	case DW_FORM_data4:
	    return read_32(p);
	/* @@@ we don't support data8 yet */
	case DW_FORM_udata:
	    return uLEB128_decode(&p);
	default:
	    fprintf(stderr,"unrecognized unsigned constant form %u\n",form);
	    exit(1);
    }
}


int dwarf_decode_flag(enum DW_FORM form,const void *p)
{
    if (form != DW_FORM_flag || *(unsigned char *) p > 1) {
	fprintf(stderr,"unrecognized or invalid flag (form %u, valud %u)\n",
	  form,*(unsigned char *) p);
	exit(1);
    }
    return *(unsigned char *) p;
}


const void *dwarf_decode_reference(enum DW_FORM form,const void *p)
{
    unsigned off;

    switch (form) {
	case DW_FORM_ref1:
	    off = *(const uint8_t *) p;
	    break;
	case DW_FORM_ref2:
	    off = read_16(p);
	    break;
	case DW_FORM_ref4:
	    off = read_32(p);
	    break;
	/* @@@ we don't support data8 yet */
	case DW_FORM_ref_udata:
	    off = uLEB128_decode(&p);
	    break;
	default:
	    fprintf(stderr,"unrecognized reference form %u\n",form);
	    exit(1);
    }
    return curr_comp_unit+off;
}


const char *dwarf_decode_string(enum DW_FORM form,const void *p)
{
    switch (form) {
	case DW_FORM_string:
	    return p;
	case DW_FORM_strp:
	    return curr_debug_str+read_32(p);
	default:
	    fprintf(stderr,"unrecognized string form %u\n",form);
	    exit(1);
    }
}


/* ----- DWARF attribute decoder ------------------------------------------- */


static enum DW_AT decode_attr(const void **abbrev,enum DW_FORM *form)
{
    enum DW_AT attr;
    enum DW_FORM tmp_form;

    attr = uLEB128_decode(abbrev);
    tmp_form = uLEB128_decode(abbrev);
    if (form) *form = tmp_form;
    return attr;
}


static void skip_attr(enum DW_FORM form,const void **buf)
{
    int size;

//printf("skip_attr in %p\n",*buf);
    if (form == DW_FORM_indirect) form = uLEB128_decode(buf);
    switch (form) {
	case DW_FORM_data1:
	case DW_FORM_ref1:
	case DW_FORM_flag:
	    (*buf)++;
	    break;
	case DW_FORM_data2:
	case DW_FORM_ref2:
	    *buf += 2;
	    break;
	case DW_FORM_data4:
	case DW_FORM_ref4:
	case DW_FORM_strp:
	    *buf += 4;
	    break;
	case DW_FORM_data8:
	case DW_FORM_ref8:
	    *buf += 8;
	    break;
	case DW_FORM_addr:
	case DW_FORM_ref_addr:
	    *buf += sizeof(void *);
	    break;
	case DW_FORM_sdata:
	    (void) sLEB128_decode(buf);
	    break;
	case DW_FORM_ref_udata:
	case DW_FORM_udata:
	    (void) uLEB128_decode(buf);
	    break;
	case DW_FORM_string:
	    while (*(*(char **) buf)++);
	    break;
	case DW_FORM_block1:
	    *buf += *(uint8_t *) *buf+1;
	    break;
	case DW_FORM_block2:
	    *buf += read_16(*buf)+1;
	    break;
	case DW_FORM_block4:
	    *buf += read_32(*buf)+1;
	    break;
	case DW_FORM_block:
	    size = uLEB128_decode(buf);
	    *buf += size;
	    break;
	default:
	    fprintf(stderr,"unrecognized form %u\n",form);
	    exit(1);
    }
//printf("skip_attr out %p\n",*buf);
}


/* ----- DWARF abbreviation table ------------------------------------------ */


static const void **abbrev_table = NULL;
static int abbrev_table_size;


static void read_abbrev_table(void)
{
    const void *p;

    if (!abbrev_table) {
	abbrev_table_size = 1024;
	abbrev_table = alloc(abbrev_table_size*sizeof(const void *));
    }
    memset(abbrev_table,0,abbrev_table_size*sizeof(const void *));
    p = curr_debug_abbrev+read_32(curr_comp_unit+6);
    while (1) {
	int num;

	num = uLEB128_decode(&p);
	if (!num) return;
	if (num >= abbrev_table_size) {
	    int new_size;

	    for (new_size = abbrev_table_size*2; new_size <= num;
	      new_size *= 2);
	    abbrev_table = realloc(abbrev_table,new_size*sizeof(const void *));
	    if (!abbrev_table) {
		perror("realloc");
		exit(1);
	    }
	    memset(abbrev_table+abbrev_table_size,0,
	      (new_size-abbrev_table_size)*sizeof(const void *));
	    abbrev_table_size = new_size;
	}
	abbrev_table[num] = p;
	(void) uLEB128_decode(&p); /* tag */
	p++; /* children */
	while (decode_attr(&p,NULL)); /* attributes */
    }
}


/* ----- DWARF info structure decoder -------------------------------------- */


#define ATTR_HASH 29


static const struct dwarf_attr *curr_attr_cache[ATTR_HASH];


const struct dwarf_attr *dwarf_attr(enum DW_AT attr)
{
    const struct dwarf_attr *walk;

    for (walk = curr_attr_cache[attr % ATTR_HASH]; walk; walk = walk->next)
	if (walk->attr == attr) break;
    return walk;
}


int dwarf_more_tags(const void *buf)
{
    return buf != curr_comp_unit_end;
}


void dwarf_skip_children(const void **buf)
{
    if (dwarf_children)
        while (dwarf_tag(buf))
            dwarf_skip_children(buf);
}


enum DW_TAG dwarf_tag(const void **buf)
{
    const char *abbrev;
    enum DW_TAG tag;
    int num,i;

    if (debug) printf("%p/%p\n",*buf,curr_comp_unit_end);
    num = uLEB128_decode(buf);
    if (!num) return 0;
    if (debug) printf("abbrev %d\n",num);
    abbrev = abbrev_table[num];
    tag = uLEB128_decode((const void **) &abbrev);
    dwarf_children = *abbrev++ == DW_CHILDREN_yes;
    if (debug)
	printf("  tag %u%s\n",tag,dwarf_children ? " (has children)" : "");
    for (i = 0; i < ATTR_HASH; i++)
	while (curr_attr_cache[i]) {
	    const struct dwarf_attr *next;

	    next = curr_attr_cache[i]->next;
	    free((void *) curr_attr_cache[i]);
	    curr_attr_cache[i] = next;
	}
    while (1) {
	struct dwarf_attr *entry;
	enum DW_AT attr;
	enum DW_FORM form;
	int hash;

	attr = decode_attr((const void **) &abbrev,&form);
if (debug) printf("    attr %u form %u\n",attr,form);
	if (!attr) break;
	hash = attr % ATTR_HASH;
	entry = alloc_t(struct dwarf_attr);
	entry->attr = attr;
	entry->form = form;
	entry->pos = *buf;
	entry->next = curr_attr_cache[hash];
	curr_attr_cache[hash] = entry;
	skip_attr(form,buf);
    }
    return tag;
}


const void *dwarf_comp_unit(void)
{
    int version;

    if (curr_comp_unit) curr_comp_unit = curr_comp_unit_end;
    else curr_comp_unit = curr_debug_info;
    if (curr_comp_unit == curr_debug_info_end) return NULL;
    version = read_16(curr_comp_unit+4);
    if (version != 2) {
	fprintf(stderr,"expected DWARF version 2, not %d\n",version);
	exit(1);
    }
    read_abbrev_table();
    curr_comp_unit_end = curr_comp_unit+read_32(curr_comp_unit)+4;
    return curr_comp_unit+11;
}


/* ----- General setup ----------------------------------------------------- */


void dwarf_setup(const void *debug_abbrev,const void *debug_info,
  const void *debug_info_end,const void *debug_str)
{
    curr_debug_abbrev = debug_abbrev;
    curr_debug_info = debug_info;
    curr_debug_info_end = debug_info_end;
    curr_debug_str = debug_str;
    curr_comp_unit = NULL;
}
