/*
 * dwxloc.c - Location expressions
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "util.h"
#include "dwarf.h"
#include "dwarfhelp.h"
#include "dwxloc.h"


#define STACK_SIZE 1024


extern int debug;


/* ----- Compute locations ------------------------------------------------- */


static unsigned long stack[STACK_SIZE];
static int sp;


#define RELOP(op) \
    assert(sp > 0); \
    stack[sp-1] = (long) stack[sp] op (long) stack[sp-1]; \
    sp--


static void run_op(const void **pc)
{
    enum DW_OP op;
    unsigned long tmp;

    op = *(unsigned char *) *pc;
    if (debug) printf("op 0x%x, pc 0x%p, sp %d\n",op,pc,sp);
    (*pc)++;
    switch (op) {
	case DW_OP_addr:
	    stack[++sp] = read_32(*pc);
	    *pc += 4;
	    break;
	case DW_OP_deref:
	    abort();
	case DW_OP_const1s:
	    stack[++sp] = *(unsigned char *) *pc;
	    if (stack[sp] & 0x80) stack[sp] |= ~0 & ~0xff;
	    (*pc)++;
	    break;
	case DW_OP_const1u:
	    stack[++sp] = *(unsigned char *) *pc;
	    (*pc)++;
	    break;
	case DW_OP_const2s:
	    stack[++sp] = read_16(*pc);
	    if (stack[sp] & 0x8000) stack[sp] |= ~0 & ~0xffff;
	    *pc += 2;
	    break;
	case DW_OP_const2u:
	    stack[++sp] = read_16(*pc);
	    *pc += 2;
	    break;
	case DW_OP_const4s:
	    stack[++sp] = read_32(*pc);
	    if (stack[sp] & 0x80000000) stack[sp] |= ~0 & ~0xffffffff;
	    *pc += 4;
	    break;
	case DW_OP_const4u:
	    stack[++sp] = read_32(*pc);
	    *pc += 4;
	    break;
/* @@@	case DW_OP_const8s: */
/* @@@	case DW_OP_const8u: */
	case DW_OP_consts:
	    stack[++sp] = sLEB128_decode(pc);
	    break;
	case DW_OP_constu:
	    stack[++sp] = uLEB128_decode(pc);
	    break;
	case DW_OP_dup:
	    assert(sp >= 0);
	    stack[sp+1] = stack[sp];
	    sp++;
	    break;
	case DW_OP_drop:
	    assert(sp >= 0);
	    sp--;
	    break;
	case DW_OP_over:
	    assert(sp > -1);
	    stack[sp+1] = stack[sp-1];
	    sp++;
	    break;
	case DW_OP_pick:
	    assert(sp >= *(unsigned char *) *pc);
	    stack[sp+1] = stack[sp-*(unsigned char *) *pc];
	    (*pc)++;
	    sp++;
	    break;
	case DW_OP_swap:
	    assert(sp > -1);
	    tmp = stack[sp];
	    stack[sp] = stack[sp-1];
	    stack[sp-1] = tmp;
	    break;
	case DW_OP_rot:
	    assert(sp > 2);
	    tmp = stack[sp];
	    stack[sp] = stack[sp-1];
	    stack[sp-1] = stack[sp-2];
	    stack[sp-2] = tmp;
	    break;
	case DW_OP_xderef:
	    abort();
	case DW_OP_abs:
	    assert(sp > -1);
	    if ((long) stack[sp] < 0) stack[sp] = -(long) stack[sp];
	    break;
	case DW_OP_and:
	    assert(sp > 0);
	    stack[sp-1] &= stack[sp];
	    sp--;
	    break;
	case DW_OP_div:
	    assert(sp > 0);
	    stack[sp-1] /= stack[sp];
	    sp--;
	    break;
	case DW_OP_minus:
	    assert(sp > 0);
	    stack[sp-1] -= stack[sp];
	    sp--;
	    break;
	case DW_OP_mod:
	    assert(sp > 0);
	    stack[sp-1] %= stack[sp];
	    sp--;
	    break;
	case DW_OP_mul:
	    assert(sp > 0);
	    stack[sp-1] *= stack[sp];
	    sp--;
	    break;
	case DW_OP_neg:
	    assert(sp >= 0);
	    stack[sp] = -stack[sp];
	    break;
	case DW_OP_not:
	    assert(sp >= 0);
	    stack[sp] = ~stack[sp];
	    break;
	case DW_OP_or:
	    assert(sp > 0);
	    stack[sp-1] |= stack[sp];
	    sp--;
	    break;
	case DW_OP_plus:
	    assert(sp > 0);
	    stack[sp-1] += stack[sp];
	    sp--;
	    break;
	case DW_OP_plus_uconst:
	    assert(sp >= 0);
	    stack[sp] += uLEB128_decode(pc);
	    break;
	case DW_OP_shl:
	    assert(sp > 0);
	    stack[sp-1] <<= stack[sp];
	    sp--;
	    break;
	case DW_OP_shr:
	    assert(sp > 0);
	    stack[sp-1] >>= stack[sp];
	    sp--;
	    break;
	case DW_OP_shra:
	    assert(sp > 0);
	    if (stack[sp] && (long) stack[sp-1] < 0)
		stack[sp-1] = (stack[sp-1] >> stack[sp]) |
		  (~0UL << (sizeof(long)*8-stack[sp-1]));
	    else stack[sp-1] >>= stack[sp];
	    sp--;
	    break;
	case DW_OP_xor:
	    assert(sp > 0);
	    stack[sp-1] ^= stack[sp];
	    sp--;
	    break;
	case DW_OP_skip:
	    pc += read_16(*pc)+2;
	    break;
	case DW_OP_bra:
	    assert(sp > 0);
	    pc += (stack[sp] ? read_16(*pc) : 0)+2;
	    sp--;
	    break;
	case DW_OP_eq:
	    RELOP(==);
	    break;
	case DW_OP_ge:
	    RELOP(>=);
	    break;
	case DW_OP_gt:
	    RELOP(>);
	    break;
	case DW_OP_le:
	    RELOP(<=);
	    break;
	case DW_OP_lt:
	    RELOP(<);
	    break;
	case DW_OP_ne:
	    RELOP(!=);
	    break;
	case DW_OP_regx:
	    abort();
	case DW_OP_fbreg:
	    abort();
	case DW_OP_bregx:
	    abort();
	case DW_OP_piece:
	    abort();
	case DW_OP_deref_size:
	    abort();
	case DW_OP_xderef_size:
	    abort();
	case DW_OP_nop:
	    break;
	case DW_OP_lo_user:
	    abort();
	case DW_OP_hi_user:
	    abort();
	default:
	    if (op >= DW_OP_lit0 && op <= DW_OP_lit31) {
		stack[++sp] = op-DW_OP_lit0;
		break;
	    }
	    if (op >= DW_OP_reg0 && op <= DW_OP_reg31) abort();
	    if (op >= DW_OP_breg0 && op <= DW_OP_breg31) abort();
	    fprintf(stderr,"unrecognized opcode %u\n",op);
	    exit(1);
    }
}


unsigned long dwx_loc_comp(const void *expr,int length,unsigned long base)
{
    const void *end = expr+length;

    stack[0] = base;
    sp = 0;
    while (expr != end)
	run_op(&expr);
    assert(!sp);
    return stack[sp];
}


unsigned long dwx_loc_here(const struct dwx_loc *list,unsigned long addr)
{
    const struct dwx_loc *walk;

    for (walk = list; walk; walk = walk->next)
	if (walk->start <= addr && walk->end > addr) {
	    const void *expr = walk->expr;
	    const void *end = expr+walk->length;

	    sp = -1;
	    while (expr != end)
		run_op(&expr);
	    assert(!sp);
	    return stack[sp];
	}
    fprintf(stderr,"location unknown for address 0x%lx\n",addr);
    exit(1);
}


/* ----- Read location list ------------------------------------------------ */


const struct dwx_loc *dwx_loc_list(enum DW_FORM form,const void *pos)
{
    struct dwx_loc *loc;
    const void *ref;

    switch (form) {
	case DW_FORM_block1:
	case DW_FORM_block2:
	case DW_FORM_block4:
	case DW_FORM_block:
	    loc = alloc_t(struct dwx_loc);
	    loc->start = 0;
	    loc->end = ~0;
	    loc->expr = dwarf_decode_block(form,pos,&loc->length);
	    loc->next = NULL;
	    return loc;
	default:
	    break;
    }
    ref = dwarf_decode_reference(form,pos);
    loc = NULL;
    while (1) {
	struct dwx_loc *new;

	new = alloc_t(struct dwx_loc);
	new->start = read_32(ref);
	ref += 4;
	new->end = read_32(ref);
	ref += 4;
	if (!new->start && !new->end) {
	    free(new);
	    break;
	}
	new->length = read_16(ref);
	ref += 2;
	new->expr = ref;
	ref += new->length;
	new->next = loc;
	loc = new;
    }
    return loc;
}
