/*
 * dwxobj.c - Read and use objects from DWARF2 information
 *
 * Written 2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <string.h>

#include "util.h"
#include "dwarf.h"
#include "dwxtype.h"
#include "dwxobj.h"


/* ----- Object lookup ----------------------------------------------------- */


const struct dwx_obj *dwx_find_unit(const struct dwx_obj *start,
  const char *path)
{
    const struct dwx_obj *obj,*best = NULL;
    int best_offset = 1 << (sizeof(int)*8-2); /* something large :-) */
    const char *basename;
    int dir_len;

    basename = strrchr(path,'/');
    if (basename) {
	dir_len = path-basename;
	basename++;
    }
    else {
	basename = path;
	dir_len = 0;
    }
    for (obj = start; obj; obj = obj->next) {
	int offset;

	if (obj->kind != dwx_obj_unit) continue;
	if (!obj->name || strcmp(obj->name,basename)) continue;
	if (!dir_len || !obj->u.unit.dir) offset = obj->u.unit.dir_len;
	else {
	    offset = obj->u.unit.dir_len-dir_len;
	    if (offset < 0) continue;
	    if (offset && obj->u.unit.dir[offset-1] != '/') continue;
	    if (!strncmp(path,obj->u.unit.dir+offset,dir_len)) continue;
	}
	if (best_offset > offset) {
	    best_offset = offset;
	    best = obj;
	    if (!offset) break;
	}
    }
    return best;
}


const struct dwx_obj *dwx_find_item(const struct dwx_obj *start,
  const char *name)
{
    const struct dwx_obj *obj,*unit;

    for (obj = start; obj; obj = obj->next) {
	if (obj->kind == dwx_obj_unit) continue;
	if (obj->name && !strcmp(obj->name,name)) return obj;
    }
    if (start->parent) return dwx_find_item(start->parent,name);
    for (unit = start; unit ; unit = unit->next) {
	if (unit->kind != dwx_obj_unit) continue;
	if (unit->children == start) continue; /* we've already done that one */
	for (obj = unit->children; obj; obj = obj->next) {
	    if (obj->kind == dwx_obj_unit) continue;
	    if (!obj->external) continue;
	    if (obj->name && !strcmp(obj->name,name)) return obj;
	}
    }
    return NULL;
}


/* ----- Input from DWARF2 ------------------------------------------------- */


static struct dwx_obj *read_objects(const struct dwx_obj *parent,
  const void **unit)
{
    struct dwx_obj *list = NULL,*last = NULL;

    while (1) {
	const void *start;
	const struct dwx_type **type = NULL;
	struct dwx_obj *obj;
	const struct dwarf_attr *attr;
	enum DW_TAG tag;
	int more;

//fprintf(stderr,"@%p\n",*unit);
	if (!parent && !dwarf_more_tags(*unit)) break;
	start = *unit;
	tag = dwarf_tag(unit);
//fprintf(stderr,"    [%p] (%d)\n",parent,tag);
	if (!tag) break;
	obj = alloc_t(struct dwx_obj);
	attr = dwarf_attr(DW_AT_name);
	obj->name = attr ? dwarf_attr_string(attr) : NULL;
//fprintf(stderr,"        %p \"%s\"\n",obj,obj->name);
	switch (tag) {
	    case DW_TAG_compile_unit:
		obj->kind = dwx_obj_unit;
		/* @@@ check language, naming, etc. */
		attr = dwarf_attr(DW_AT_comp_dir);
		obj->u.unit.dir = attr ? dwarf_attr_string(attr) : NULL;
		if (!obj->u.unit.dir) obj->u.unit.dir_len = 0;
		else {
		    obj->u.unit.dir_len = strlen(obj->u.unit.dir);
		    if (obj->u.unit.dir_len &&
		      obj->u.unit.dir[obj->u.unit.dir_len-1] == '/')
			obj->u.unit.dir_len--;
		}
		break;
	    case DW_TAG_subprogram:
	    case DW_TAG_inlined_subroutine:
		obj->kind = dwx_obj_fn;
		attr = dwarf_attr(DW_AT_return_addr);
		obj->u.fn.ret_addr = attr ? dwx_loc_list(attr->form,attr->pos)
		  : NULL;
		attr = dwarf_attr(DW_AT_frame_base);
		obj->u.fn.frame_addr = attr ?
		  dwx_loc_list(attr->form,attr->pos) : NULL;
		type = &obj->u.fn.type;
		break;
	    case DW_TAG_lexical_block:
		obj->kind = dwx_obj_block;
		break;
	    case DW_TAG_variable:
	    case DW_TAG_formal_parameter:
	    case DW_TAG_constant:
		obj->kind = dwx_obj_var;
		attr = dwarf_attr(DW_AT_location);
		obj->u.var.loc = attr ? dwx_loc_list(attr->form,attr->pos) :
		  NULL;
		attr = dwarf_attr(DW_AT_const_value);
		obj->u.var.constant = attr ? dwarf_attr_block(attr,NULL) : NULL;
		type = &obj->u.var.type;
		break;
	    case DW_TAG_label:
		obj->kind = dwx_obj_label;
		break;
	    default:
		free(obj);
		*unit = start; /* @@@ quite inefficient */
		(void) read_type(unit);
		continue;
	}
	attr = dwarf_attr(DW_AT_external);
	obj->external = attr ? dwarf_attr_flag(attr) : 0;
	attr = dwarf_attr(DW_AT_low_pc);
	obj->low_pc = attr ? (unsigned long) dwarf_attr_addr(attr) : 0;
	attr = dwarf_attr(DW_AT_start_scope);
	if (attr) obj->low_pc += dwarf_attr_uconst(attr);
	attr = dwarf_attr(DW_AT_high_pc);
	obj->high_pc = attr ? (unsigned long) dwarf_attr_addr(attr) : 0;
	more = dwarf_children;
	if (type) *type = resolve_type();
	obj->parent = parent;
	if (last) last->next = obj;
	else list = obj;
	obj->children = more ? read_objects(obj,unit) : NULL;
	obj->next = NULL;
	last = obj;
    }
    return list;
}


struct dwx_obj *dwx_read_objects(const void *unit)
{
    return read_objects(NULL,&unit);
}
