/*
 * dwxtype.c - Read types from DWARF2 information
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "util.h"
#include "dwarf.h"
#include "dwxout.h"
#include "dwxloc.h"
#include "dwxtype.h"


#define TYPE_HASH 4111
static struct dwx_type *type_hash[TYPE_HASH];

#define ENUM_HASH 7001
static struct dwx_enum_member *enum_hash[ENUM_HASH];


static const struct dwx_type last_child;
  /* end marker (returned by read_type) */


/* ----- Type lookup ------------------------------------------------------- */


const struct dwx_type *dwx_lookup_type(enum dwx_kind kind,const char *name,
  int size,int skip_n)
{
    int i;
    const struct dwx_type *walk;

    for (i = 0; i < TYPE_HASH; i++)
	for (walk = type_hash[i]; walk; walk = walk->next) {
	    if (walk->kind != kind) continue;
	    if (name && (!walk->name || strcmp(name,walk->name))) continue;
	    if (size != -1 && walk->size != -1 && size != walk->size) continue;
	    if (!skip_n--) return walk;
	}
    return NULL;

}


/* ----- Helper functions -------------------------------------------------- */


static int hash_enum(const char *s)
{
    int hash = 0;

    while (*s)
	hash = (128*hash+*s++) % ENUM_HASH;
    return hash;
}


static const struct dwarf_attr *require(enum DW_AT attr_id)
{
    const struct dwarf_attr *attr;

    attr = dwarf_attr(attr_id);
    if (attr) return attr;
    fprintf(stderr,"required attribute %u not found\n",attr_id);
    exit(1);
}


/* ----- Allocation and hashing -------------------------------------------- */


static struct dwx_type *get_type(unsigned long def)
{
    struct dwx_type *type;
    int hash;

    if (!def) return NULL;
    hash = def % TYPE_HASH;
    for (type = type_hash[def % TYPE_HASH]; type; type = type->next)
	if (type->def == def) break;
    return type;
}


static struct dwx_type *new_type(enum dwx_kind kind,unsigned long def)
{
    struct dwx_type *type;
    const struct dwarf_attr *attr;
    int hash;

    if (get_type(def)) {
	fprintf(stderr,"new_type: type %lu already exists\n",def);
	exit(1);
    }
    type = alloc_t(struct dwx_type);
    type->kind = kind;
    attr = dwarf_attr(DW_AT_name);
    type->name = attr ? dwarf_attr_string(attr) : NULL;
    attr = dwarf_attr(DW_AT_byte_size);
    type->size = attr ? dwarf_attr_uconst(attr) : -1;
    type->def = def;
    hash = def % TYPE_HASH;
    type->next = type_hash[hash];
    type_hash[hash] = type;
    return type;
}


/* ----- Resolve types ----------------------------------------------------- */


/*
 * Warning: resolve_type will clobber the attribute cache, so call it only
 * after all other attributes have been retrieved.
 */


const struct dwx_type *resolve_type(void)
{
    const struct dwarf_attr *attr;
    const void *ref;
    const struct dwx_type *type;

    attr = dwarf_attr(DW_AT_type);
    if (!attr) return NULL; /* void */
    ref = dwarf_attr_reference(attr);
    type = get_type((unsigned long) ref);
    if (type) return type;
    type = read_type(&ref);
    if (type) return type;
    fprintf(stderr,"failed to resolve type at %p\n",ref);
    exit(1);
}


/* ----- Type parsers ------------------------------------------------------ */


static const struct dwx_type *array_type(const void **p,unsigned long def)
{
    struct dwx_type *type;
    const struct dwx_type **last_type = (const struct dwx_type **) &type;
    const struct dwarf_attr *attr;
    int stride_size = -1;

    assert(dwarf_children);
    attr = dwarf_attr(DW_AT_stride_size);
    if (attr) stride_size = dwarf_attr_uconst(attr);
    *last_type = resolve_type();
    while (1) {
	enum DW_TAG tag;
	struct dwx_type *new;

	tag = dwarf_tag(p);
	if (!tag) break;
	new = new_type(dwx_array,def);
	new->u.array.stride_size = stride_size;
	stride_size = -1;
	new->u.array.type = *last_type;
	*last_type = new;
	last_type = &new->u.array.type;
	switch (tag) {
	    case DW_TAG_subrange_type:
		attr = dwarf_attr(DW_AT_upper_bound);
		if (attr) {
		    new->u.array.elements = dwarf_attr_uconst(attr)+1;
		    break;
		}
		attr = dwarf_attr(DW_AT_count);
		if (attr) {
		    new->u.array.elements = dwarf_attr_uconst(attr);
		    break;
		}
		new->u.array.elements = -1; /* unspecified */
		break;
	    case DW_TAG_enumeration_type:
		abort();
	    default:
		abort();
	}
	def = 0;
    }
    return type;
}


static const struct dwx_type *struct_type(enum dwx_kind kind,const void **p,
  unsigned long def)
{
    struct dwx_type *type;
    struct dwx_struct_member **last;

    type = new_type(kind,def);
    type->u._struct.members = NULL;
    if (!dwarf_children) return type;
    last = &type->u._struct.members;
    while (1) {
	enum DW_TAG tag;
	const struct dwarf_attr *attr;

	tag = dwarf_tag(p);
	if (!tag) break;
	assert(tag == DW_TAG_member);
	*last = alloc_t(struct dwx_struct_member);
	(*last)->name = dwarf_attr_string(require(DW_AT_name));
	attr = dwarf_attr(DW_AT_byte_size);
	(*last)->size = attr ? dwarf_attr_uconst(attr) : -1;
	attr = dwarf_attr(DW_AT_data_member_location);
	if (!attr) (*last)->offset = 0;
	else {
	    const void *expr;
	    int length;

	    expr = dwarf_attr_block(attr,&length);
	    /*
	     * This isn't entirely correct, because the location expression
	     * might produce different results if the base is non-zero
	     * (e.g. for alignment). @@@
	     *
	     * Cleaner solution: pass the current offset along while
	     * resolving nested types.
	     */
	    (*last)->offset = dwx_loc_comp(expr,length,0);
	}
	(*last)->type = resolve_type();
	last = &(*last)->next;
	*last = NULL;
    }
    return type;
}


static const struct dwx_type *enum_type(const void **p,unsigned long def)
{
    struct dwx_type *type;
    struct dwx_enum_member **last;

    type = new_type(dwx_enum,def);
    type->u._enum.members = NULL;
    if (!dwarf_children) return type;
    last = &type->u._enum.members;
    while (1) {
	enum DW_TAG tag;
	int hash;

	tag = dwarf_tag(p);
	if (!tag) break;
	assert(tag == DW_TAG_enumerator);
	*last = alloc_t(struct dwx_enum_member);
	(*last)->name = dwarf_attr_string(require(DW_AT_name));
	(*last)->value = dwarf_attr_sconst(require(DW_AT_const_value));
	hash = hash_enum((*last)->name);
	(*last)->next_hash = enum_hash[hash];
	enum_hash[hash] = *last;
	last = &(*last)->next;
	*last = NULL;
    }
    return type;
}


static const struct dwx_type *type_modifier(enum dwx_kind kind,
  unsigned long def)
{
    struct dwx_type *type;

    type = new_type(kind,def);
    type->u.modifier.type = resolve_type();
    return type;
}


static const struct dwx_type *subroutine_type(unsigned long def)
{
    struct dwx_type *type;

    type = new_type(dwx_fn,def);
    type->u.fn.type = resolve_type();
    return type;
}


static const struct dwx_type *typedef_type(unsigned long def)
{
    struct dwx_type *type;

    type = new_type(dwx_typedef,def);
    type->u._typedef.type = resolve_type();
    return type;
}


static const struct dwx_type *base_type(unsigned long def)
{
    enum DW_ATE enc;
    struct dwx_type *type;

    enc = dwarf_attr_uconst(require(DW_AT_encoding));
    switch (enc) {
	case DW_ATE_signed:
	case DW_ATE_signed_char:
	    type = new_type(dwx_sint,def);
	    break;
	case DW_ATE_unsigned:
	case DW_ATE_unsigned_char:
	    type = new_type(dwx_uint,def);
	    break;
	case DW_ATE_float:
	    type = new_type(dwx_float,def);
	    break;
	default:
	    fprintf(stderr,"unknown encoding %u\n",enc);
	    exit(1);
    }
    return type;
}


/* ----- Type synthesis ---------------------------------------------------- */


static struct dwx_type *alloc_type(enum dwx_kind kind,int size)
{
    struct dwx_type *new;

    new = alloc_t(struct dwx_type);
    new->kind = kind;
    new->name = NULL;
    new->size = size;
    new->def = 0;
    return new;
}


const struct dwx_type *dwx_basic_type(enum dwx_kind kind,int size)
{
    return alloc_type(kind,size);
}


const struct dwx_type *dwx_add_ptr(const struct dwx_type *type)
{
    struct dwx_type *new;

    new = alloc_type(dwx_ptr,4); /* @@@ fixme */
    new->u.modifier.type = type;
    return new;
}


const struct dwx_type *dwx_add_array(const struct dwx_type *type,
  int elements,int stride_size)
{
    struct dwx_type *new;

    /*
     * This algorithm isn't very good. E.g. struct { int x; char y; }; may have
     * a size of 5, but has a natural alignment of 4 (assuming 32 bit ints), so
     * the stride size would still be 8. gcc typically includes the alignment
     * in the size, but that doesn't mean relying on it is safe.
     */
    new = alloc_type(dwx_array,
      (stride_size == -1 ? dwx_sizeof(type) : stride_size)*elements);
    new->u.array.elements = elements;
    new->u.array.type = type;
    new->u.array.stride_size = stride_size;
    return new;
}


const struct dwx_type *dwx_add_fn(const struct dwx_type *type)
{
    struct dwx_type *new;

    new = alloc_type(dwx_fn,4);
    new->u.fn.type = type;
    return new;
}


const struct dwx_type *dwx_index(const struct dwx_type *type,
  unsigned long *offset,int index)
{
    assert(type->kind == dwx_array);
    if (type->u.array.elements != -1 &&
      (index < 0 || index >= type->u.array.elements)) {
	fprintf(stderr,"index %d on array [%d]\n",index,
	  type->u.array.elements);
	exit(1);
    }
    *offset += index*(type->u.array.stride_size != -1 ?
      type->u.array.stride_size : dwx_sizeof(type->u.array.type));
    return type->u.array.type;
}


const struct dwx_type *dwx_member(const struct dwx_type *type,
  unsigned long *offset,const char *name)
{
    const struct dwx_struct_member *m;

    assert(type->kind == dwx_struct || type->kind == dwx_union);
    for (m = type->u._struct.members; m; m = m->next)
	if (!strcmp(m->name,name)) break;
    if (!m) {
	fprintf(stderr,"no member \"%s\"\n",name);
	exit(1);
    }
    *offset += m->offset;
    return m->type;
}


int dwx_lookup_enum(const char *name,int *value)
{
    const struct dwx_enum_member *walk;
    int hits = 0;

    for (walk = enum_hash[hash_enum(name)]; walk; walk = walk->next_hash)
	if (!strcmp(walk->name,name))
	    switch (hits) {
		case 0:
		    *value = walk->value;
		    hits++;
		    break;
		case 1:
		    if (*value == walk->value) break;
		    return 2;
	    }
    return hits;
}


/* ----- Main loop --------------------------------------------------------- */


const struct dwx_type *read_type(const void **p)
{
    unsigned long def;
    const struct dwx_type *type;

    def = (unsigned long) *p;
    type = get_type(def);
    if (type) {
	dwarf_tag(p);
	dwarf_skip_children(p);
	return type;
    }
    switch (dwarf_tag(p)) {
	case DW_TAG_array_type:
	    return array_type(p,def);
	case DW_TAG_enumeration_type:
	    return enum_type(p,def);
	case DW_TAG_structure_type:
	    return struct_type(dwx_struct,p,def);
	case DW_TAG_subroutine_type:
	    return subroutine_type(def);
	case DW_TAG_typedef:
	    return typedef_type(def);
	case DW_TAG_union_type:
	    return struct_type(dwx_union,p,def);
	case DW_TAG_base_type:
	    return base_type(def);
	case DW_TAG_pointer_type:
	    return type_modifier(dwx_ptr,def);
	case DW_TAG_const_type:
	    return type_modifier(dwx_const,def);
	case DW_TAG_volatile_type:
	    return type_modifier(dwx_volatile,def);
	case 0:
	    return &last_child;
	default:
	    return NULL;
    }
}


void dwx_read_types(const void *unit)
{
    while (dwarf_more_tags(unit))
	read_type(&unit);
}
