/*
 * elf.c - Read DWARF2 debugging information from ELF file
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */

/*
 * References:
 * http://www.skyfree.org/linux/references/ELF_Format.pdf
 * http://www.eagercon.com/dwarf/dwarf3std.htm
 */


#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <elf.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include "dwarf.h"


/* ----- Map a file -------------------------------------------------------- */


const void *mmap_file(const char *name,int *size)
{
    int fd;
    struct stat st;
    const void *start;

    fd = open(name,O_RDONLY);
    if (fd < 0) {
	perror(name);
	exit(1);
    }
    if (fstat(fd,&st) < 0) {
	perror(name);
	exit(1);
    }
    start = mmap(NULL,st.st_size,PROT_READ,MAP_PRIVATE,fd,0);
    if (start == MAP_FAILED) {
	perror("mmap");
	exit(1);
    }
    if (size) *size = st.st_size;
    return start;
}


/* ----- ELF structure access ---------------------------------------------- */


static void read_elf_types(const void *elf)
{
    const Elf32_Ehdr *Ehdr;
    int i;
    const void *debug_abbrev = NULL,*debug_info = NULL,*debug_str = NULL;
    const void *debug_info_end;
    const void *pos,*strings;

    Ehdr = elf;
    if (memcmp(Ehdr->e_ident,ELFMAG,SELFMAG)) {
	fprintf(stderr,"unrecognized ELF file format\n");
	exit(1);
    }
    if (Ehdr->e_ident[EI_CLASS] != ELFCLASS32) {
	fprintf(stderr,"we only support 32 bit ELF, sorry\n");
	exit(1);
    }
    if (Ehdr->e_ident[EI_DATA] != ELFDATA2LSB) {
	fprintf(stderr,"we only support little-endian ELF, sorry\n");
	exit(1);
    }
    if (!Ehdr->e_shoff) {
	fprintf(stderr,"file has no section header table\n");
	exit(1);
    }
    if (!Ehdr->e_shstrndx) {
	fprintf(stderr,"file has no section name string table\n");
	exit(1);
    }
    pos = elf+Ehdr->e_shoff;
    strings = elf+
      ((const Elf32_Shdr *) (pos+Ehdr->e_shentsize*Ehdr->e_shstrndx))->
      sh_offset;
    for (i = 0; i < Ehdr->e_shnum; i++) {
	const Elf32_Shdr *Shdr = pos;
	const char *name = strings+Shdr->sh_name;

	if (!strcmp(name,".debug_abbrev")) debug_abbrev = elf+Shdr->sh_offset;
	if (!strcmp(name,".debug_info")) {
	    debug_info = elf+Shdr->sh_offset;
	    debug_info_end = debug_info+Shdr->sh_size;
	}
	if (!strcmp(name,".debug_str")) debug_str = elf+Shdr->sh_offset;
	pos += Ehdr->e_shentsize;
    }
    if (!debug_abbrev) {
	fprintf(stderr,".debug_abbrev section not found\n");
	exit(1);
    }
    if (!debug_info) {
	fprintf(stderr,".debug_info section not found\n");
	exit(1);
    }
    dwarf_setup(debug_abbrev,debug_info,debug_info_end,debug_str);
}


void open_elf(const char *name)
{
    read_elf_types(mmap_file(name,NULL));
}
